/***************************************************************************
                          lunacal.cpp  -  description
                             -------------------
    Open Source Code. See main.cpp for details.
    copyright            : (C) 2013 by Holtkamp
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <QtGui>

#include "lunacal.h"
#include "lunamph.h"
#include "astrolib.h"

#include "lndatedialog.h"
#include "lnttutdialog.h"
#include "lnobsdialog.h"
#include "lnholidaydlg.h"
#include "lunquickdlg.h"
#include "lndaydialog.h"
#include "lnmoondialog.h"
#include "lnyeardialog.h"
#include "lnmapdialog.h"
#include "lnsolardialog.h"
// define the windows input and output globaly

WoutData *outdata = new WoutData;
WinpData *indata = new WinpData;
int fntsz = 10;  // display font size
double cang = 100.0;  // cutoff angle for yearly total

LunacalApp::LunacalApp()
{
  setWindowTitle(tr("LunaCal - 5.6"));
  setWindowIcon(QPixmap(":/luncal32.png"));

  QLocale::setDefault(QLocale::English);

  createActions();
  createMenus();
  createToolBars();
  createStatusBar();
  resize(751, 510);

  initView();
  setFocus();

  slotViewStatusBar(false);

  outdata->curlines = 0;
  lwelcm(indata, outdata);  // initialize data and display welcoming message

  view->showtxt(false, outdata);
  view->setMap(indata->eclbuf.emagrid, indata->eclbuf.emagrey, indata->eclbuf.cstep,
               indata->mapprj, indata->centmer);
}

 LunacalApp::~LunacalApp()
{
}

/** initializes all QActions of the application */
void LunacalApp::createActions(){

  QPixmap saveIcon, dateIcon, locationIcon, phaseIcon, calendarIcon, eclipseIcon, holidayIcon;
  saveIcon = QPixmap(":/filesave.png");
  dateIcon = QPixmap(":/luncdate.png");
  locationIcon = QPixmap(":/luncloc.png");
  phaseIcon = QPixmap(":/luncphas.png");
  calendarIcon = QPixmap(":/luncldsp.png");
  eclipseIcon = QPixmap(":/luncecl.png");
  holidayIcon = QPixmap(":/luncholy.png");

  fileSave = new QAction(QIcon(saveIcon), tr("&Save File"), this); 
  fileSave->setShortcut(tr("Ctrl+F"));
  fileSave->setStatusTip(tr("Saves the Moon phases / Calendar"));
  fileSave->setWhatsThis(tr("Save File.\n\nSaves the Moon phases table and the calendar"
          " on the selected (or default) output file."));
  connect(fileSave, SIGNAL(triggered()), this, SLOT(slotFileSave()));

  fileSaveAs = new QAction(tr("Save File &As"), this);
  fileSaveAs->setStatusTip(tr("Saves the Moon phases / Calendar under a new filename"));
  fileSaveAs->setWhatsThis(tr("Save As\n\nSaves the the Moon phases table and the calendar"
          " on a newly specified output file."));
  connect(fileSaveAs, SIGNAL(triggered()), this, SLOT(slotFileSaveAs()));

  fileQuit = new QAction(tr("E&xit"), this);
  fileQuit->setShortcut(tr("Ctrl+Q"));
  fileQuit->setStatusTip(tr("Quits LunaCal"));
  fileQuit->setWhatsThis(tr("Exit\n\nQuits LunaCal"));
  connect(fileQuit, SIGNAL(triggered()), this, SLOT(slotFileQuit()));

  inputDate = new QAction(QIcon(dateIcon), tr("&Date"), this);   
  inputDate->setShortcut(tr("Ctrl+D"));
  inputDate->setStatusTip(tr("Select year and timezone"));
  inputDate->setWhatsThis(tr("Date\n\nSelects the year and the timezone for the calculations."));
  connect(inputDate, SIGNAL(triggered()), this, SLOT(slotInputDate()));

  inputDeltaTime = new QAction(tr("Delta-&Time"), this);
  inputDeltaTime->setShortcut(tr("Ctrl+T"));
  inputDeltaTime->setStatusTip(tr("Set the Delta-Time (difference TT - UTC)"));
  inputDeltaTime->setWhatsThis(tr("Delta-Time\n\nSets the difference between Terrestial Time (TT)"
            " and Universial Time (UTC). If you don't know what this is use the default."));
  connect(inputDeltaTime, SIGNAL(triggered()), this, SLOT(slotInputDeltaTime()));

  inputLocation = new QAction(QIcon(locationIcon), tr("&Location"), this);
  inputLocation->setShortcut(tr("Ctrl+L"));
  inputLocation->setStatusTip(tr("Set the location."));
  inputLocation->setWhatsThis(tr("Location\n\nSets the location for which the local circumstances are calculated."
        " The coordinates can be in decimal degrees or in DDD.MMss format (Degrees.MinutesSeconds)"));
  connect(inputLocation, SIGNAL(triggered()), this, SLOT(slotInputLocation()));
 
  calcPhases = new QAction(QIcon(phaseIcon), tr("&Moon Phases"), this);
  calcPhases->setShortcut(tr("Ctrl+M"));
  calcPhases->setStatusTip(tr("Calculate the phases of the Moon"));
  calcPhases->setWhatsThis(tr("Moon Phases\n\nCalculate the phases of the Moon for the specified year."));
  connect(calcPhases, SIGNAL(triggered()), this, SLOT(slotCalcPhases()));

  calcCalendar = new QAction(QIcon(calendarIcon), tr("&Calendar"), this);
  calcCalendar->setShortcut(tr("Ctrl+C"));
  calcCalendar->setStatusTip(tr("Display the calendar"));
  calcCalendar->setWhatsThis(tr("Calendar\n\nDisplay the calendar for the specified year."
	"\nThe Julian Calendar will be used before the year 1582"));
  connect(calcCalendar, SIGNAL(triggered()), this, SLOT(slotCalcCalendar()));

  calcEclipses = new QAction(QIcon(eclipseIcon), tr("&Eclipses"), this);
  calcEclipses->setShortcut(tr("Ctrl+E"));
  calcEclipses->setStatusTip(tr("Calculate lunar and solar eclipses"));
  calcEclipses->setWhatsThis(tr("Eclipses\n\nCalculates the lunar and solar eclipses for the specified year."
            " A world map showing the solar eclipse visibilities can be displayed."));
  connect(calcEclipses, SIGNAL(triggered()), this, SLOT(slotCalcEclipses()));

  calcHolidays = new QAction(QIcon(holidayIcon), tr("H&oliday"), this);
  calcHolidays->setShortcut(tr("Ctrl+O"));
  calcHolidays->setStatusTip(tr("Calculate cultural and religious holidays"));
  calcHolidays->setWhatsThis(tr("Holidays\n\nCalculates for the specified year the following holidays:"
        "\n\nAsh Wednesday\nEaster\n\nJewish New Year\nYom Kippur"
        "\n\nIslamic New Year\nFirst Day of Ramadan\n\nIndian New Year\n\nChinese New Year"));
  connect(calcHolidays, SIGNAL(triggered()), this, SLOT(slotCalcHolidays()));

  calcDay = new QAction(tr("D&ay Details"), this);
  calcDay->setShortcut(tr("Ctrl+A"));
  calcDay->setStatusTip(tr("Display details of the day"));
  calcDay->setWhatsThis(tr("Day Details\n\nShows the weekday, day of the year, week of the year"
        " as well as corresponding date in the Jewish, Islamic, Indian and Chinese calendar of the"
        " specified date"));
  connect(calcDay, SIGNAL(triggered()), this, SLOT(slotCalcDay()));

  calcSky = new QAction(tr("&Sky Details"), this);
  calcSky->setShortcut(tr("Ctrl+S"));
  calcSky->setStatusTip(tr("Display Sun and Moon details of the day"));
  calcSky->setWhatsThis(tr("Sky Details\n\nShows the Julian Date, position of the Sun"
        " as well as the position, phase and libration of the Moon for the specified date and time."));
  connect(calcSky, SIGNAL(triggered()), this, SLOT(slotCalcSky()));

  calcYearTot = new QAction(tr("&Year Totals"), this);
  calcYearTot->setShortcut(tr("Ctrl+Y"));
  calcYearTot->setStatusTip(tr("Display total hours of Sun- and Moonshine and twilight"));
  calcYearTot->setWhatsThis(tr("Year Totals\n\nShows the total hours of Sun and Moon at the specified"
        " location and the specified minimum angle."));
  connect(calcYearTot, SIGNAL(triggered()), this, SLOT(slotCalcYearTot()));

  calcSolarPanel = new QAction(tr("Sola&r Panel"), this);
  calcSolarPanel->setShortcut(tr("Ctrl+R"));
  calcSolarPanel->setStatusTip(tr("Display maximum percentage and electricity of a fixed solar panel"));
  calcSolarPanel->setWhatsThis(tr("Solar Panel\n\nShows how a solar panel with a fixed azimuth and elevation at the given location compares (in percent) with one which tracks the sun. Also given is the maximum amount of energy in kWh which would be produced by each square meter of such a panel in one year if there were now clouds."));
  connect(calcSolarPanel, SIGNAL(triggered()), this, SLOT(slotCalcSolarPanel()));

  calcSatEcl = new QAction(tr("Spec&ial Eclipses"), this);
  calcSatEcl->setShortcut(tr("Ctrl+I"));
  calcSatEcl->setStatusTip(tr("Display special solar eclipse map"));
  calcSatEcl->setWhatsThis(tr("Special Eclipses\n\nDisplays the centerline of a solar eclipse around a"
            " specified New Moon date onto a surface which is above the Earth surface by specified height."
            " This enables to check for eclipses at the height of satellite orbits." ));
  connect(calcSatEcl, SIGNAL(triggered()), this, SLOT(slotCalcSatEcl()));

  viewMapGrid = new QAction(tr("Ecli&pse Settings"), this);
  viewMapGrid->setShortcut(tr("Ctrl+P"));
  viewMapGrid->setStatusTip(tr("Select world map display and time steps for eclipses"));
  viewMapGrid->setWhatsThis(tr("Eclipse Settings\n\nSpecifies the type of eclipse world map"
        " as well as how coordinates should be drawn on the map."
        " Also specify the time steps to be used for the central eclipse listings."));
  connect(viewMapGrid, SIGNAL(triggered()), this, SLOT(slotViewMapGrid()));

  viewFontSize = new QAction(tr("Font Si&ze"), this);
  viewFontSize->setShortcut(tr("Ctrl+Z"));
  viewFontSize->setStatusTip(tr("Set the font size for the screen display"));
  viewFontSize->setWhatsThis(tr("Font Size\n\nSpecifies the font size (8 or 10) for the screen display."));
  connect(viewFontSize, SIGNAL(triggered()), this, SLOT(slotViewFontSize()));

  viewToolBar = new QAction(tr("Tool&bar"), this);
  viewToolBar->setStatusTip(tr("Enables/disables the toolbar"));
  viewToolBar->setWhatsThis(tr("Toolbar\n\nEnables/disables the toolbar"));
  viewToolBar->setCheckable(true);
  connect(viewToolBar, SIGNAL(toggled(bool)), this, SLOT(slotViewToolBar(bool)));

  viewStatusBar = new QAction(tr("&Statusbar"), this);
  viewStatusBar->setStatusTip(tr("Enables/disables the statusbar"));
  viewStatusBar->setWhatsThis(tr("Statusbar\n\nEnables/disables the statusbar"));
  viewStatusBar->setCheckable(true);
  connect(viewStatusBar, SIGNAL(toggled(bool)), this, SLOT(slotViewStatusBar(bool)));

  helpAboutApp = new QAction(tr("&About..."), this);
  helpAboutApp->setStatusTip(tr("About the application"));
  helpAboutApp->setWhatsThis(tr("About\n\nAbout the application"));
  connect(helpAboutApp, SIGNAL(triggered()), this, SLOT(slotHelpAbout()));

  aboutQtAct = new QAction(tr("About &Qt"), this);
  aboutQtAct->setStatusTip(tr("Show the Qt library's About box"));
  connect (aboutQtAct, SIGNAL(triggered()), qApp, SLOT(aboutQt()));

  quickHelp = new QAction(tr("&Help..."), this);
  quickHelp->setStatusTip(tr("Quick help"));
  quickHelp->setWhatsThis(tr("Help\n\nA summary help of how to use LunaCal."));
  connect(quickHelp, SIGNAL(triggered()), this, SLOT(slotQuickHelp()));

  whts = QWhatsThis::createAction(this);
  whts->setStatusTip(tr("What's This?"));

}

void LunacalApp::createMenus()
{
  fileMenu = menuBar()->addMenu(tr("&File"));
  fileMenu->addAction(fileSave);
  fileMenu->addAction(fileSaveAs);
  fileMenu->addSeparator();
  fileMenu->addAction(fileQuit);

  inputMenu = menuBar()->addMenu(tr("&Input"));
  inputMenu->addAction(inputDate);
  inputMenu->addAction(inputDeltaTime);
  inputMenu->addSeparator();
  inputMenu->addAction(inputLocation);

  calcMenu = menuBar()->addMenu(tr("Calc&ulate"));
  calcMenu->addAction(calcPhases);
  calcMenu->addAction(calcCalendar);
  calcMenu->addAction(calcEclipses);
  calcMenu->addAction(calcHolidays);
  calcMenu->addSeparator();
  calcMenu->addAction(calcDay);
  calcMenu->addAction(calcSky);
  calcMenu->addAction(calcYearTot);
  calcMenu->addAction(calcSolarPanel);
  calcMenu->addSeparator();
  calcMenu->addAction(calcSatEcl);

  viewMenu = menuBar()->addMenu(tr("&View"));
  viewMenu->addAction(viewMapGrid);
  viewMenu->addAction(viewFontSize);
  viewMenu->addSeparator();
  viewMenu->addAction(viewToolBar);
  viewMenu->addAction(viewStatusBar);

  helpMenu = menuBar()->addMenu(tr("&Help"));
  helpMenu->addAction(helpAboutApp);
  helpMenu->addAction(aboutQtAct);
  helpMenu->addSeparator();
  helpMenu->addAction(quickHelp);
  helpMenu->addSeparator();
  helpMenu->addAction(whts);
}

void LunacalApp::createToolBars()
{
  fileToolBar = addToolBar(tr("LunaCal Toolbar"));
  fileToolBar->addAction(fileSave);
  fileToolBar->addSeparator();
  fileToolBar->addAction(inputDate);
  fileToolBar->addAction(inputLocation);
  fileToolBar->addSeparator();
  fileToolBar->addAction(calcPhases);
  fileToolBar->addAction(calcCalendar);
  fileToolBar->addAction(calcEclipses);
  fileToolBar->addAction(calcHolidays);
  fileToolBar->addSeparator();
  fileToolBar->addSeparator();
  fileToolBar->addAction(whts);
}

void LunacalApp::createStatusBar()
{
  statusBar()->showMessage(tr("Ready."), 2000);
}

void LunacalApp::initView()
{
  // set the main widget here

 imageLabel = new QLabel;
 imageLabel->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Ignored);
 imageLabel->setScaledContents(true);

 view=new LunacalView(imageLabel);

 sv = new QScrollArea();
 sv->setWidget(view);
 setCentralWidget(sv); 

}

/////////////////////////////////////////////////////////////////////
// SLOT IMPLEMENTATION
/////////////////////////////////////////////////////////////////////

void LunacalApp::keyPressEvent(QKeyEvent *e)
{
 int n;

 n = 0;  
 setFocus();

 switch (e->key()) {
 case Qt::Key_Down:
          view->scrollDsp(0, 10, indata);
          break;
 case Qt::Key_Left:
          view->scrollDsp(10, 0, indata);
          break;
 case Qt::Key_Up:
          view->scrollDsp(0, -10, indata);
          break;
 case Qt::Key_Right:
          view->scrollDsp(-10, 0, indata);
          break;
 case '1':
          n = 1;
          break;
 case '2':
          n = 2;
          break;
 case '3':
          n = 3;
          break;
 case '4':
          n = 4;
          break;
 case '5':
          n = 5;
          break;
 case '6':
          n = 6;
          break;
 case '7':
          n = 7;
          break;
 case '8':
          n = 8;
          break;
 case '9':
          n = 9;
          break;
 case '0':
          n = 10;
          break;
 case 'a':
 case 'A':
          slotCalcDay();
          break;
 case 'c':
 case 'C':
          slotCalcCalendar();
          break;
 case 'd':
 case 'D':
          slotInputDate();
          break;
 case 'e':
 case 'E':
          slotCalcEclipses();
          break;
 case 'f':
 case 'F':
          slotFileSave();
          break;
 case 'h':
 case 'H':
          slotQuickHelp();
          break;
 case 'l':
 case 'L':
          slotInputLocation();
          break;
 case 'm':
 case 'M':
          slotCalcPhases();
          break;
 case 'o':
 case 'O':
          slotCalcHolidays();
          break;
 case 'p':
 case 'P': 
          slotViewMapGrid();
          break;
 case 'r':
 case 'R':
          slotCalcSolarPanel();
          break;
 case 's':
 case 'S':
          slotCalcSky();
          break;
 case 't':
 case 'T':
          slotInputDeltaTime();
          break;
 case 'y':
 case 'Y':
          slotCalcYearTot();
          break;
 case 'z':
 case 'Z':
          slotViewFontSize();
          break;

 default: QWidget::keyPressEvent(e);
 };

 if (n > 0) view->keyPress(n);         
}

void LunacalApp::slotFileSave()
{
  statusBar()->showMessage(tr("Writing Lunar Phases / Calendar..."));

  FILE* otptr;
  char otxt[255];
  int j;

  indata->filsflg = 1;
  outdata->curlines = 0;
  moonph(indata, outdata);

  indata->filsflg = 2;
  otptr = fopen (indata->svfname,"a");
  if (otptr != NULL)
   {
    fprintf(otptr,
           "\n\n\n\t\tHolidays for %4i\n\t\t=================\n\n",indata->year);
    GetEaster(indata->year, otxt);
    fputs(otxt,otptr);
    GetJewish(indata, otxt);
    fputs(otxt,otptr);
    GetIslam(indata->year, otxt);
    fputs(otxt,otptr);
    GetIndia(indata->year, otxt);
    fputs(otxt,otptr);
    GetChina(indata, outdata, otxt);
    fputs(otxt,otptr);
    fprintf(otptr,"\n\nJewish and Islamic dates are tabular dates");
    fprintf(otptr,"\nand begin at sunset on the previous evening\n");

    outdata->curlines = 0;
    caldisp(indata, outdata);
    fprintf(otptr,"\n\n\n");
    for (j=0; j<38; j++)
     {
      fputs(outdata->wbuf[j],otptr);
      fprintf(otptr,"\n");
     };
    fclose(otptr);
   };

  indata->filsflg = 0;

  outdata->curlines = 0;
  if (indata->lstcall <= 1) moonph(indata, outdata);
  if (indata->lstcall == 2) caldisp(indata, outdata);
  view->showtxt(true, outdata);
  indata->invalflg = 1;

  statusBar()->showMessage(tr("Ready."));
}

void LunacalApp::slotFileSaveAs()
{
  const char* ga;
  int j, k;
  
  statusBar()->showMessage(tr("Writing Lunar Phases / Calendar to a new file..."));
  QString fn = QFileDialog::getSaveFileName(this, 0, 0);

  if (!fn.isEmpty())
  {
    ga = fn.toAscii();
    k = strlen(ga);
    if (k > 0)
     {
      if (k > 54) k = 54;
      for (j=0; j<k; j++) indata->svfname[j] = ga[j];
      indata->svfname[54] = '\0'; // just in case
     };
   slotFileSave();  
  }
  else
  {
    statusBar()->showMessage(tr("Operation aborted"), 2000);
  }

  statusBar()->showMessage(tr("Ready."));
}

void LunacalApp::slotFileQuit()
{
  statusBar()->showMessage(tr("Exiting application..."));

  putcfg(indata, 0);
  
  qApp->quit();

  statusBar()->showMessage(tr("Ready."));
}

void LunacalApp::slotInputDate()
{
  statusBar()->showMessage(tr("Input the year..."));

  LnDateDialog *dlg = new LnDateDialog();
  dlg->setParams(indata->year, indata->tzone);
  if (dlg->exec())
   {
    indata->year = dlg->getYear();
    indata->tzone = dlg->getTzone();
    if (indata->del_auto) indata->del_tdut = DefTdUt(indata->year);
   }
  else
   {
    statusBar()->showMessage(tr("Input cancelled."));
    return;
   };

  statusBar()->showMessage(tr("Ready."));
}

void LunacalApp::slotInputDeltaTime()
{
  double dfdut;
  statusBar()->showMessage(tr("Input Delta-Time..."));

  dfdut = DefTdUt(indata->year);
  LnTtutDialog *dlg = new LnTtutDialog();
  dlg->setParams(indata->del_tdut, dfdut, indata->del_auto);

  if (dlg->exec())
   {
    indata->del_tdut = dlg->getDelut();
    indata->del_auto = dlg->getDelauto();
   }
  else
   {
    statusBar()->showMessage(tr("Input cancelled."));
    return;
   };

  statusBar()->showMessage(tr("Ready."));
}

void LunacalApp::slotInputLocation()
{
  double lat, lng, height;
  int j, k;
  QString gn;
  QByteArray ba;
  bool ddmm;
  
  statusBar()->showMessage(tr("Input the location..."));

  LnObsDialog *dlg = new LnObsDialog();
  dlg->setParams(indata->geolat, indata->geolong, indata->geoheight, indata->geoname, indata->dmsflg );
  if (dlg->exec())
   {
    dlg->getParams(lat, lng, height, ddmm );
    indata->geolat = lat;
    indata->geolong = lng;
    indata->geoheight = height;
    indata->dmsflg = ddmm;

    gn = dlg->getGeoname();
    ba = gn.toAscii();
    k = ba.size()+1;
    
    char *ga = new char[k];
 
    strcpy (ga, ba.data());
   
    k = strlen(ga);
    if (k > 0)
     {
      if (k > 30) k = 30;
      for (j=0; j<30; j++) indata->geoname[j] = ' ';  // to clear everything
      for (j=0; j<k; j++) indata->geoname[j] = ga[j];
      indata->geoname[30] = '\0'; // just in case
     };
    delete[] ga;
   }
  else
   {
    statusBar()->showMessage(tr("Input cancelled."));
    return;
   };

  statusBar()->showMessage(tr("Ready."));
}

void LunacalApp::slotCalcPhases()
{
  statusBar()->showMessage(tr("Calculate the phases of the Moon..."));

  outdata->curlines = 0;
  moonph(indata, outdata);
  indata->lstcall = 1;
  view->showtxt(true, outdata);

  statusBar()->showMessage(tr("Ready."));
}

void LunacalApp::slotCalcCalendar()
{
  statusBar()->showMessage(tr("Display the Calendar..."));

  outdata->curlines = 0;
  caldisp(indata, outdata);
  indata->lstcall = 2;
  view->showtxt(true, outdata);

  statusBar()->showMessage(tr("Ready."));
}

void LunacalApp::slotCalcEclipses()
{
  statusBar()->showMessage(tr("Calculate lunar and solar eclipses..."));

  indata->lstcall = 3;
  view->showEcltxt(indata);
  
  statusBar()->showMessage(tr("Click respective line to get eclipse details."));
}

void LunacalApp::slotCalcHolidays()
{
	char otxt[255];
  QString s;

  statusBar()->showMessage(tr("Calculate cultural and religious holidays..."));

  LnHolidayDialog *dlg = new LnHolidayDialog();

  sprintf(otxt,
    "\t\tHolidays for %4i\n\t\t==============\n\n",indata->year);
  s = otxt;
  dlg->setHolidayLine(s);

  GetEaster(indata->year, otxt);
  s = otxt;
  dlg->setHolidayLine(s);

  GetJewish(indata, otxt);
  s = otxt;
  dlg->setHolidayLine(s);
 
  GetIslam(indata->year, otxt);
  s = otxt;
  dlg->setHolidayLine(s);
  
  GetIndia(indata->year, otxt);
  s = otxt;
  dlg->setHolidayLine(s);
 
  GetChina(indata, outdata, otxt);
  s = otxt;
  dlg->setHolidayLine(s);
  
  sprintf(otxt,"\n\n\nJewish and Islamic dates are tabular dates");
  s = otxt;
  dlg->setHolidayLine(s);
  
  sprintf(otxt,"\nand begin at sunset on the previous evening");
  s = otxt;
  dlg->setHolidayLine(s);

  dlg->setHolidayLine("$$$");
  dlg->exec();

  statusBar()->showMessage(tr("Ready."));
}

void LunacalApp::slotCalcDay()
{
  int year, month, day;
  
  statusBar()->showMessage(tr("Calculate day details..."));

  LnDayDialog *dlg = new LnDayDialog();
  year = indata->year;
  month = indata->month;
  day = indata->day;
  dlg->setParams(year, month, day, fntsz);
  if (dlg->exec())
   {
    dlg->getParams(year, month, day);
    if (year != indata->year)
     {
      if (indata->del_auto) indata->del_tdut = DefTdUt(year);
     };
    indata->year = year;
    indata->month = month;
    indata->day = day;
   };
  
  statusBar()->showMessage(tr("Ready."));
}

void LunacalApp::slotCalcSky()
{
  statusBar()->showMessage(tr("Calculate sky details..."));

  LnMoonDialog *dlg = new LnMoonDialog();

  dlg->setParams(indata->year, indata->month, indata->day, indata->hour, indata->minute, indata->second,
       indata->tzone, indata->del_tdut, indata->del_auto, indata->geolat, indata->geolong, indata->geoheight,
       indata->geoname, indata->dmsflg, fntsz );
  if (dlg->exec())
   {
    dlg->getParams(indata->year, indata->month, indata->day, indata->hour, indata->minute, indata->second,
         indata->tzone, indata->del_tdut );
   };

  statusBar()->showMessage(tr("Ready."));
}

void LunacalApp::slotCalcYearTot()
{
  char otxt[550];
  bool ok;
  QString s3;
  
  statusBar()->showMessage(tr("Calculate yearly totals..."));

  LnYearDialog *dlg = new LnYearDialog();

  cang = QInputDialog::getDouble(this,"Yearly Totals",
         "Cutoff Angle (deg - if > 90 standard output):",
                 cang, -90.0, 100.0, 2, &ok);
  if (ok)
   {
    CompYearTot (indata, cang, otxt);
    s3 = otxt;
  
    dlg->setParams(s3, fntsz);
    dlg->exec();
   }; 

  statusBar()->showMessage(tr("Ready."));
}

void LunacalApp::slotCalcSolarPanel()
{
  char otxt[550];
  double spaz, spel, speff;
  int spm1, spm2;
  QString s3;
  
  statusBar()->showMessage(tr("Calculate solar panel statistics"));

  LnSolarDialog *dlg = new LnSolarDialog();

  spaz = indata->spnaz;
  spel = indata->spnel;
  speff = indata->spneff;
  spm1 = indata->spnm1;
  spm2 = indata->spnm2;
  
  dlg->setParams(spaz, spel, speff, spm1, spm2);
  if (dlg->exec())
   {
    dlg->getParams(spaz, spel, speff, spm1, spm2);
    indata->spnaz = spaz;
    indata->spnel = spel;
    indata->spneff = speff;
    indata->spnm1 = spm1;
    indata->spnm2 = spm2;

    LnYearDialog *dlg2 = new LnYearDialog();

    CompYearTot (indata, 12345.6, otxt);
    s3 = otxt;
  
    dlg2->setParams(s3, fntsz);
    dlg2->exec();
   }; 

  statusBar()->showMessage(tr("Ready."));
}

void LunacalApp::slotCalcSatEcl()
{
  statusBar()->showMessage(tr("Show special solar eclipse map..."));

  indata->lstcall = 4;

  view->showSatEcl(indata);

  statusBar()->showMessage(tr("Ready."));
}

void LunacalApp::slotViewMapGrid()
{
  int mgrid, mgrey, mstep, mapprj;
  double centmer;
  
  statusBar()->showMessage(tr("Specify the eclipse map and time steps..."));

  LnMapDialog *dlg = new LnMapDialog();

  mgrid = indata->eclbuf.emagrid;
  mgrey = indata->eclbuf.emagrey;
  mstep = indata->eclbuf.cstep;
  mapprj = indata->mapprj;
  centmer = indata->centmer;
  
  dlg->setParams(mgrid, mgrey, mstep, mapprj, centmer);
  if (dlg->exec())
   {
    dlg->getParams(mgrid, mgrey, mstep, mapprj, centmer);
    view->setMap(mgrid, mgrey, mstep, mapprj, centmer);
    indata->eclbuf.emagrid = mgrid;
    indata->eclbuf.emagrey = mgrey;
    indata->eclbuf.cstep = mstep;
    indata->mapprj = mapprj;
    indata->centmer = centmer;
   };
  
  statusBar()->showMessage(tr("Ready."));
}

void LunacalApp::slotViewFontSize()
{
  bool ok;
  int res;
  
  statusBar()->showMessage(tr("Specify the font size..."));

  res = QInputDialog::getInteger(this,"Font Size", "Specify font size for display", fntsz, 8, 12, 2, &ok);
  if (ok)
   {
    fntsz = res; 
    view->setFont(fntsz);
   };
    
  statusBar()->showMessage(tr("Ready."));
}

void LunacalApp::slotViewToolBar(bool toggle)
{
  statusBar()->showMessage(tr("Toggle toolbar..."));
  // turn Toolbar on or off

  if (toggle== false)
  {
    fileToolBar->hide();
  }
  else
  {
    fileToolBar->show();
  };

  statusBar()->showMessage(tr("Ready."));
}

void LunacalApp::slotViewStatusBar(bool toggle)
{
  statusBar()->showMessage(tr("Toggle statusbar..."));
  //turn Statusbar on or off

  if (toggle == false)
  {
    statusBar()->hide();
  }
  else
  {
    statusBar()->show();
  }

  statusBar()->showMessage(tr("Ready."));
}

void LunacalApp::slotHelpAbout()
{
  QMessageBox::about(this,tr("About..."),
             tr("Lunacal\nVersion 5.6\n(c) February 2013\n by Gerhard HOLTKAMP") );
}

void LunacalApp::slotQuickHelp()
{
  QString s;
  s = ":/lunacalHelp.txt";
  QFile file(s);

  statusBar()->showMessage(tr("Provide general help with LunaCal"));

  LunQuickDialog *dlg = new LunQuickDialog();

  if(!file.open(QIODevice::ReadOnly | QIODevice::Text))
   {
    statusBar()->showMessage(tr("*** ERROR ***  unable to open Help file"));
    return;
   };

   QTextStream in(&file);

   while (!in.atEnd())
    {
     s = in.readLine();
     dlg->setHelpLine(s);
     dlg->setHelpLine("\n");
    };

  dlg->setHelpLine("$$$");
  dlg->exec();

  statusBar()->showMessage(tr("Ready."));

}

