/***************************************************************************
* Calculate the Phases of the Moon, Eclipses and Calendars                 *
*                                                                          *
* INPUT : Year and time zone                                               *
* OUTPUT: List of dates and times (Universal Time) of                      *
*	  the phases of the Moon (New, Full, first and last quarter.       *
*         This list will be displayed on screen and also written on        *
*         a file with the name "lunphase.txt".                             *
*                                                                          *
* Note: The accuracy of the calculation is about 1 or 2 minutes if staying *
*       within 100 years or so of the year 2000. A standard difference of  *
*       of 65 sec between Dynamical Time and Universal Time has been       *
*       assumed which is about correct for the year 2000. If you take      *
*       times well beyond that date you would have to allow for            *
*       corresponding corrections. (The correction is done in function     *
*       ckphase below. Subtracted from td is the correction in fraction    *
*       of a day.)                                                         *
*                                                                          *
* The algorithm for the phases of the Moon and of the dates for the        *
* equinoxes and solstices of the Sun as well as the dates of eclipses      *
* are based on Jean Meeus "Astronomical Formulae for Calculators",         *
* Willman-Bell, Inc., Richmond, Virginia, 1988                             *
*                                                                          *
* The algorithm for the Modified Julian Date and the calendar as well as   *
* the detailed eclipse calculations are based on                           *
* O.Montebruck and T.Pfleger, "Astronomy with a PC",                       *
* Springer Verlag, Berlin, Heidelberg, New York, 1989                      *
*                                                                          *
* The algorithms for the various calendric holidays are based on the       *
* "Explanatory Supplement to the Astronomical Almanac"                     *
*
* Linux C++ version                                                        *
*    Open Source Code. See main.cpp for details.
* Author: Gerhard HOLTKAMP,        10-FEB-2013
***************************************************************************/

/*------------ include files and definitions -----------------------------*/
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cmath>
#include <ctime>
using namespace std;

#include "lunamph.h"
#include "astrolib.h"
#include "nolocal.h"

#define MAXLUN 15

FILE *outptr;  /* Output File */

typedef double PMJD[MAXLUN];  /* array of MJD's of the respective phase */

/* const double PI = 3.14159265359; */
const double degrad = M_PI / 180.0;

double atan23 (double y, double x)
 {
  /* redefine atan2 so that it does'nt crash when both x and y are 0 */
  double result;

  if ((x == 0) && (y == 0)) result = 0;
  else result = atan2 (y, x);

  return result;
 }

/*---------------------- function GetMonth --------------------------------*/
void GetMonth (int mm, char* mchr)
 {
  // mm is the number of the month
  // mchr is a char[4] array to receive the three letters of the month

  switch (mm)
   {
    case 1 : strcpy(mchr,"JAN"); break;
    case 2 : strcpy(mchr,"FEB"); break;
    case 3 : strcpy(mchr,"MAR"); break;
    case 4 : strcpy(mchr,"APR"); break;
    case 5 : strcpy(mchr,"MAY"); break;
    case 6 : strcpy(mchr,"JUN"); break;
    case 7 : strcpy(mchr,"JUL"); break;
    case 8 : strcpy(mchr,"AUG"); break;
    case 9 : strcpy(mchr,"SEP"); break;
    case 10 : strcpy(mchr,"OCT"); break;
    case 11 : strcpy(mchr,"NOV"); break;
    case 12 : strcpy(mchr,"DEC"); break;
    default: strcpy(mchr,"ERR");
   };
}

/*---------------------- function phmjd --------------------------------*/

double phmjd (double yearf, double phase, double tdut,
				  int& eph, double& ejd, double& emag)
  /*
    Calculate the Modified Julian Date of the occurrence of the specified
    phase of the Moon and check for possible eclipses.
    yearf : year.fraction of date close to the Moon phase
    phase : 0    for New Moon,
            0.25 for First Quarter,
            0.5  for Full Moon,
            0.75 for Last Quarter.
    tdut = TDT - UT in sec

    RETURN: the MJD of the lunar phase event

    OUTPUT:
    eph : phase of the eclipse on that date
          0 = no eclipse, 1 = partial solar eclipse,
          2 = non-central annular, 3 = non-central total
          4 = annular solar, 5 = total solar,
          6 = annular/total solar, -1 =partial penumbral lunar,
          -2 = total penumbral lunar, -3 = partial umbral lunar,
          -4 = total umbral lunar
    ejd : Modified Julian Date of the maximum of the eclipse (if any)
    emag : magnitude of the eclipse (if any)
  */
 {
  double tt, jd, k, m, p, f;
  double s, c, gam, u;
  int tst;
  Eclipse eclp;

  // preliminary (modified) Julian Date 
  k = floor ((yearf - 1900.0) * 12.3685) + phase;
  tt = k / 1236.85;
  jd = 166.56 + (132.87 - 0.009173*tt)*tt;
  jd = 15020.25933 + 29.53058868*k
                   + (0.0001178 - 0.000000155*tt)*tt*tt
                   + 0.00033 * sin (degrad*jd);

  eph = 0;   // assume no eclipse

  // Sun's mean anomaly 
  m = degrad * (359.2242 + 29.10535608*k
               - (0.0000333 - 0.00000347*tt)*tt*tt);
  // Moon's mean anomaly 
  p = degrad * (306.0253 + 385.81691808*k
              + (0.0107306 + 0.00001236*tt)*tt*tt);
  // 2* Moon's argument of latitude 
  f = 2.0 * degrad * (21.2964 + 390.67050646*k
                   - (0.0016528 - 0.00000239*tt)*tt*tt);

  if ((phase == 0) || (phase == 0.5))  // for Full and New Moon 
   { 
    k = (0.1734 - 0.000393*tt) * sin (m)
        + 0.0021 * sin (2.0 * m)
        - 0.4068 * sin (p)
        + 0.0161 * sin (2.0 * p)
        - 0.0004 * sin (3.0 * p)
        + 0.0104 * sin (f)
        - 0.0051 * sin (m + p)
        - 0.0074 * sin (m - p)
        + 0.0004 * sin (f + m)
        - 0.0004 * sin (f - m)
        - 0.0006 * sin (f + p)
        + 0.0010 * sin (f - p)
        + 0.0005 * sin (m + 2.0*p);

    // check for possible eclipses.
    f = f / 2.0;
    if (fabs(sin(f)) <= 0.36)   // eclipses are possible
     {
      ejd = (0.1734 - 0.000393*tt) * sin(m)
            + 0.0021 * sin(2.0*m)
            - 0.4068 * sin(p)
            + 0.0161 * sin(2.0*p)
            - 0.0051 * sin(m+p)
            - 0.0074 * sin(m-p)
            - 0.0104 * sin(2.0*f);
      ejd += jd;

      s = 5.19595 - 0.0048*cos(m) + 0.0020*cos(2.0*m) - 0.3283*cos(p)
                  - 0.0060*cos(m+p) + 0.0041*cos(m-p);

      c = 0.2070*sin(m) + 0.0024*sin(2.0*m) - 0.0390*sin(p)
                        + 0.0115*sin(2.0*p) - 0.0073*sin(m+p)
                        - 0.0067*sin(m-p) + 0.0117*sin(2.0*f);

      gam = s*sin(f) + c*cos(f);
      u = 0.0059 + 0.0046*cos(m) - 0.0182*cos(p)
                 + 0.0004*cos(2.0*p) - 0.0005*cos(m+p);

      if (phase == 0)          // check for solar eclipse
       {
        if (fabs(gam) <= (1.5432+u))
         {
          if (fabs(gam) < 0.9972)    // central eclipse
           {
            emag = 1.0;
            if (u < 0) eph = 5;  // total eclipse
            else
             {
              if (u > 0.0047) eph = 4;  // annular eclipse
              else
               {
                if (u < (0.00464*cos(asin(gam)))) eph = 6; // annular/total
                else eph = 4;  // annular
               }; // end if u>0.0047
             }; // end if u<0
           }  // end if fabs(gam) if
          else      // non-central eclipse
           {
            eph = 1;   // assume partial eclipse
            if (fabs(gam) < (0.9972+fabs(u))) // non-central annular or total
             {
              eph = eclp.solar(ejd,tdut,s,c);
              emag = 1.0;
             };
            if (eph == 1)     // get magnitude of partial eclipse
                emag = (1.5432 + u - fabs(gam)) / (0.5460 + 2.0*u);
            if (emag < 0.025) // check if low mag eclipse is OK
             {
              eph = 0;
              u = 1.0 / 720; // 2 min steps 
              for (int j=0; j < 288; j++)
               {
                tt = ejd - 0.2 + double(j)*u;
                tst = eclp.solar(tt,tdut,s,c);
                if (tst > 0) eph = tst;
               };
             };

           }
         }
       }  // end of solar eclipse check

      if (phase == 0.5)        // check for lunar eclipse
       {
        c = (1.5572 + u - fabs(gam)) / 0.5450;
        if (c > 0)
         {
          s = (1.0129 - u - fabs(gam)) / 0.5450;
          if (s < 0)   // penumbral eclipse
           {
            emag = c;
            if (emag > 1) eph = -2;
            else eph = -1;
           }
          else         // umbral eclipse
           {
            emag = s;
            if (emag > 1) eph = -4;
            else eph = -3;
           }
         }
       }
     }
   };

  if ((phase == 0.25) || (phase == 0.75)) // for first and last quarter 
   { 
    k = (0.1721 - 0.0004*tt) * sin (m)
        + 0.0021 * sin (2.0 * m)
        - 0.6280 * sin (p)
        + 0.0089 * sin (2.0 * p)
        - 0.0004 * sin (3.0 * p)
        + 0.0079 * sin (f)
        - 0.0119 * sin (m + p)
        - 0.0047 * sin (m - p)
        + 0.0003 * sin (f + m)
        - 0.0004 * sin (f - m)
        - 0.0006 * sin (f + p)
        + 0.0021 * sin (f - p)
        + 0.0003 * sin (m + 2.0*p)
        + 0.0004 * sin (m - 2.0*p)
        - 0.0003 * sin (2.0*m + p);
    if (phase == 0.25)
     { 
      k += 0.0028 - 0.0004*cos(m) + 0.0003*cos(p);
     };
    if (phase == 0.75)
     { 
      k += -0.0028 + 0.0004*cos(m) - 0.0003*cos(p);
     };
   };

  jd = jd + k;

  return jd;
 }

/*----------------------- function ckphase ---------------------------*/
void ckphase (double minmjd, double maxmjd, double yr, double timezone,
              double deltdut, int &mp, PMJD p, double phase, WinpData* inp)
  /* calculate the respective phase for one (MJD-) date and check whether
     this date is within the limits given by minmjd and maxmjd.
     Use yr as the year.fraction close to the desired date.
     If the date is within the limits store the result in the
     respective array.
     Also check for possible occurences of eclipses and store the
     respective data in an array
  */
 {
  double td, ejd, emag;
  int j, eph;

  td = phmjd (yr, phase, deltdut*86400.0, eph, ejd, emag);
  // correct for timezone and difference between UT and TDT 
  td = td + timezone/24.0;
  td = td - deltdut;  // correct for difference between TDT and UT.

  // check whether the date is within the respective year 

  if ((td >= minmjd) &&  (td <= maxmjd) && (mp < MAXLUN))
   {
    if (mp==0)
     {
      p[0]=td;
      mp=1;
     }
    else
     {
      if (p[mp-1] < (td - 0.1))
       {
        p[mp]=td;
        mp=mp+1;
       };
     };
   };

  // now the eclipses 
  if (eph != 0)
   {
    td = ejd + timezone/24.0;
    td = td - deltdut;  // correct for difference between TDT and UT.

    // check whether the date is within the respective year 
    if ((td >= minmjd)&&(td <= maxmjd)&&(inp->eclbuf.numecl<GBL_ECLBUF))
     {
      if (inp->eclbuf.numecl == 0)
       {
        inp->eclbuf.eclmjd[0] = td;
        inp->eclbuf.magnitude[0] = emag;
        inp->eclbuf.phase[0] = eph;
        inp->eclbuf.numecl = 1;
       }
      else
       {
        for (j=0; j<inp->eclbuf.numecl; j++)
         {          // check whether the date is already stored
          if (fabs(inp->eclbuf.eclmjd[j]- td) < 0.01) eph = 0;
         };
        if (eph != 0)
         {
          j = inp->eclbuf.numecl;
          inp->eclbuf.eclmjd[j] = td;
          inp->eclbuf.magnitude[j] = emag;
          inp->eclbuf.phase[j] = eph;
          inp->eclbuf.numecl += 1;
         };
       }
     }
   }
 }

/*----------------------- function dtmstr -----------------------------*/
void dtmstr(double jdmoon, char *dts)
  /* 
    Convert the Modified Julian Date jdmoon into a corresponding string
    *dts which has the format "MMM DD HH:MM"
  */
{
  int dd, mm, yy, deg, mnt;
  double hh, sec;
  char mchr[4];

  // convert MJD into corresponding date and time 
  caldat(jdmoon, dd, mm, yy, hh);
  dms (hh,deg,mnt,sec);
  if (sec>30.0) {mnt++;};
  if (mnt>59)
   {
    deg++;
    mnt=0;
   };

  // store data in their positions 
  GetMonth (mm, mchr);
  dts[0]=mchr[0];
  dts[1]=mchr[1];
  dts[2]=mchr[2];
  dts[3]=' ';
  sprintf(&dts[4],"%2i %2i:%02i", dd, deg, mnt);
  dts[12]=' ';
}

void dtmstrT(double jdmoon, char *dts)
  // Same as dtmstr but put a tab between the date and time
 {
  dtmstr(jdmoon, dts);
  dts[6] = '\t';
  dts[12] = '\t';
 }
  
/*----------------------- function getout ----------------------------*/
void getout(double maxmjd, PMJD p0, PMJD p25, PMJD p5, PMJD p75,
            int mp0, int mp25, int mp5, int mp75,
            int &j0, int &j25, int &j5, int &j75, char *outbuf)
 /*
   Assemble one line of output which contains the times for New Moon,
   First Quarter, Full Moon and Last Quarter.
   Make sure that the times are in ascending order.
   The counters j0, j25, j5 and j75 are updated correspondingly.
   Each date/time entry has the format "MMM dd hh:mm".
   The position for the entry for New Moon start at 0,
   First Quarter at 15, Full Moon at 31 and Last Quarter at 47.
  */
{
  int i;
  double mnew, mfirst, mfull, mlast;

  // set buffer blank 
  for (i=0;i<60;i++) outbuf[i] = ' ';

  // Assign the latest times 
  if (j0<mp0) mnew=p0[j0];
  else mnew=maxmjd+1;
  if (j25<mp25) mfirst=p25[j25];
  else mfirst=maxmjd+1;
  if (j5<mp5) mfull=p5[j5];
  else mfull=maxmjd+1;
  if (j75<mp75) mlast=p75[j75];
  else mlast=maxmjd+1;

  // check which times have to be outputted 
  if ((mnew<mfirst) && (mnew<mfull) && (mnew<mlast))
   {
    if (mnew <= maxmjd)
     {
      dtmstr(mnew,&outbuf[0]);
      j0=j0+1;
     };
   };
  if ((mfirst<mfull) && (mfirst<mlast))
   {
    if (mfirst <= maxmjd)
     {
      dtmstr(mfirst,&outbuf[15]);
      j25=j25+1;
     };
   };
  if ((mfull<mlast))
   {
    if (mfull <= maxmjd)
     {
      dtmstr(mfull,&outbuf[31]);
      j5=j5+1;
     };
   };
  if (mlast <= maxmjd)
   {
    dtmstr(mlast,&outbuf[47]);
    j75=j75+1;
   };
  outbuf[60]='\0';
}

/*--------------------- Function PutNxtLn -------------------------------*/
void PutNxtLn(WoutData* outdata, char *otxt)  // copy line into buffer
 {
  int i, j;
  int maxlen = GBL_MAXCPL;

  i = outdata->curlines;

  if (outdata->curlines < GBL_MAXLIN)
   {
    for (j = 0; j<maxlen; j++) outdata->wbuf[i][j] = otxt[j];
    outdata->curlines++;
   };
 }

void PutNxtLnT(WoutData* outdata, char *otxt)  // copy line into buffer and mark for Tab
 {
  int i, j;
  int maxlen = GBL_MAXCPL;

  i = outdata->curlines;

  if (outdata->curlines < GBL_MAXLIN)
   {
    for (j = 0; j<maxlen; j++) outdata->wbuf[i][j] = otxt[j];
    outdata->wbuf[i][maxlen-1] = '~'; 
    outdata->curlines++;
   };
 }

/*--------------------- Fuction PutSpace ----------------------------------*/
void PutSpace(WoutData* outdata, int nspace = 1)  // write "nspace" space lines
 {
  char otxt[GBL_MAXCPL] = "   ";

  for (int i = 0; i<nspace; i++) PutNxtLn(outdata, otxt);
 }

/*----------------------- function putphase ---------------------------*/
void putphase (int year, double timezone, double maxmjd, PMJD p0, PMJD p25,
               PMJD p5, PMJD p75, int mp0, int mp25, int mp5, int mp75,
               WoutData* outdata)
 // Output the table of Lunar phases for the respective year.
{
  char outbuf[61];
  int j0, j25, j5, j75;

  sprintf(outbuf,"> Timezone %5.2lf <", timezone);
  decdot(outbuf);
  PutNxtLn (outdata,outbuf);
  sprintf(outbuf,"          Phases of the Moon for %4i", year);
  PutNxtLnT (outdata,outbuf);
  sprintf(outbuf,"          ===========================");
  PutNxtLnT (outdata,outbuf);
  PutSpace(outdata);
  sprintf (outbuf,"  New Moon    First Quarter     Full Moon      Last Quarter");
  PutNxtLnT (outdata,outbuf);
  PutSpace(outdata);

  if (outptr != NULL)
   {
    sprintf(outbuf,"\n+++> Timezone %5.2lf <+++", timezone);
    decdot(outbuf);
    fprintf(outptr,outbuf);
    fprintf(outptr,"\n\n\n\t\tPhases of the Moon for %4i", year);
    fprintf(outptr,"\n\t\t===========================\n");
    fprintf(outptr,
        "\n  New Moon    First Quarter     Full Moon      Last Quarter\n\n");
   };

  j0=j25=j5=j75=0;

  while ((j0<mp0) || (j25<mp25) || (j5<mp5) || (j75<mp75))
   {
    getout(maxmjd,p0,p25,p5,p75,mp0,mp25,mp5,mp75,j0,j25,j5,j75,
                     outbuf);
    PutNxtLnT (outdata,outbuf);

    if (outptr != NULL)
     {
      outbuf[59]='\n';
      fputs(outbuf,outptr);
     };
   };
 }

/*------------------------- solstice ---------------------------------*/

void solstice(int year, double timezone, double deltut, WoutData* outdata)
 {
  /* get solstices and equinoxes for the year and output the result
     in buffer outdata accurate to the minute.
     Timezone is the respective difference to UT,
     deltut is the difference TDT - UT.
  */

  const double pi4 = M_PI/2.0;
  const double abrt = 1.58186934368e-7;  // time-delay due to aberation
  const int maxit = 12;

  double jd_eqx[4];  // MJD of equinox and solstice dates
  double y, ls, t;
  char dts[13];
  char outbuf[127];
  int k, j;
  Mat3 mx;
  Vec3 ps;

  Sun200 sun;

  y = year / 1000.0;
  for (k=0; k<4; k++)
   {
    switch (k)      // get preliminary values
     {
      case 0: jd_eqx[0] = 365.2421376*year - 678861.2145
                          + (0.0679190 - 0.0027879*y)*y*y; // March Eqx
                          break;
      case 1: jd_eqx[1] = 365.2417284*year - 678767.2514
                          + (0.0093320*y - 0.0530180)*y*y; // Jun Solst.
                          break;
      case 2: jd_eqx[2] = 365.2425055*year - 678674.8022
                          + (0.0019401*y - 0.1266890)*y*y; // Sep Eqx
                          break;
      case 3: jd_eqx[3] = 365.2428898*year - 678586.108
                          - (0.0109650 + 0.0084885*y)*y*y; // Dec Solst.
                          break;
     };

    // now the iteration for the exact values
    t = julcent(jd_eqx[k]);
    mx = nutecl (t, ls);

    j = 0;
    do
     {
      t = julcent(jd_eqx[k])-abrt;
      ps = sun.position(t);
      ps = mxvct(mx, ps);
      ps = carpol(ps);
      ls = ps[1];
      ls = 58.1313429644 * sin(k*pi4 - ls);
      jd_eqx[k] += ls;
      j+=1;
     } while ((ls > 1.157e-5) && (j < maxit));

    jd_eqx[k] += timezone/24.0;
    jd_eqx[k] -= deltut;  // correct for difference between TDT and UT.
   }

  outdata->wsolstice = jd_eqx[3];  // needed for the Chinese New Year

  PutSpace(outdata,2);
  sprintf(outbuf,"          The Sun for %4i", year);
  PutNxtLnT (outdata,outbuf);
  sprintf(outbuf,"          ================");
  PutNxtLnT (outdata,outbuf);
  PutSpace(outdata);
  if (outptr != NULL)
   {
    fprintf(outptr,"\n\n\t\tThe Sun for %4i", year);
    fprintf(outptr,"\n\t\t================\n");
   };

  strcpy(outbuf,"  Equinoxes \t");
  dtmstr(jd_eqx[0],dts);
  dts[12] = '\0';
  strcat (outbuf,dts);
  strcat (outbuf,"  \t");
  dtmstr(jd_eqx[2],dts);
  dts[12] = '\0';
  strcat (outbuf,dts);
  PutNxtLn (outdata,outbuf);
  if (outptr != NULL)
   {
    outbuf[59]='\n';
    fputs(outbuf,outptr);
   };

  strcpy(outbuf,"  Solstices \t");
  dtmstr(jd_eqx[1],dts);
  dts[12] = '\0';
  strcat (outbuf,dts);
  strcat (outbuf,"  \t");
  dtmstr(jd_eqx[3],dts);
  dts[12] = '\0';
  strcat (outbuf,dts);

  PutNxtLn (outdata,outbuf);

  if (outptr != NULL)
   {
    outbuf[59]='\n';
    fprintf(outptr,"\n");
    fputs(outbuf,outptr);
   };
 }

/*------------------------- putecl ---------------------------------*/

void putecl(int year, WinpData* indata, WoutData* outdata)
 {
  // output the eclipse dates in buffer outdata accurate to the minute.

  char dts[13];
  char outbuf[127];
  char magbuf[30];
  int j, p;

  PutSpace(outdata,2);
  sprintf(outbuf,"          ECLIPSES for %4i", year);
  PutNxtLnT (outdata,outbuf);
  sprintf(outbuf,"          =================");
  PutNxtLnT (outdata,outbuf);
  PutSpace(outdata);

  if (outptr != NULL)
   {
    fprintf(outptr,"\n\n\n\t\tEclipses for %4i", year);
    fprintf(outptr,"\n\t\t=================\n");
   };

  for (j=0; j<indata->eclbuf.numecl; j++)
   {
    dtmstrT(indata->eclbuf.eclmjd[j],dts);
    dts[12] = '\0';
    strcpy (outbuf,dts);
    p = indata->eclbuf.phase[j];
    switch (p)
     {
      case 1: strcat(outbuf,"\t Partial eclipse of the Sun");
              sprintf(magbuf,"  (magnitude:%5.2f)",indata->eclbuf.magnitude[j]);
              decdot0(magbuf);
              strcat(outbuf,magbuf);
              break;

      case 2:
      case 4: strcat(outbuf,"\t Annular eclipse of the Sun");
              break;

      case 3:
      case 5: strcat(outbuf, "\t Total eclipse of the Sun");
              break;

      case 6: strcat(outbuf, "\t Annular/Total Solar Eclipse");
              break;

      case -1:
      case -2: strcat(outbuf, "\t Penumbral eclipse of the Moon");
               sprintf(magbuf,"  (magnitude:%5.2f)",indata->eclbuf.magnitude[j]);
               decdot0(magbuf);
               strcat(outbuf,magbuf);
               break;

      case -3: strcat(outbuf, "\t Partial eclipse of the Moon");
               sprintf(magbuf,"  (magnitude:%5.2f)",indata->eclbuf.magnitude[j]);
               decdot0(magbuf);
               strcat(outbuf,magbuf);
               break;

      case -4: strcat(outbuf, "\t Total eclipse of the Moon");
               sprintf(magbuf,"  (magnitude:%5.2f)",indata->eclbuf.magnitude[j]);
               decdot0(magbuf);
               strcat(outbuf,magbuf);
               break;
     };
     
    PutNxtLn (outdata,outbuf);
    if (outptr != NULL)
     {
      fprintf(outptr,"\n");
      fputs(outbuf,outptr);
     };  
   }
 }

/*------------------------- moonph ---------------------------------*/

void moonph(WinpData* indata, WoutData* outdata)
{
  int mp0, mp25, mp5, mp75;  // counter of respective phase entries 
  PMJD p0, p25, p5, p75;
  int day, month, year;
  double yr, yx, hour, timezone, deltdut;
  double minmjd, maxmjd;

  // open the output file 
  if (indata->filsflg == 1)
   {
    if (indata->svfirst) outptr = fopen (indata->svfname,"w");
    else
     {
      if (indata->svfapnd)
       {
        outptr = fopen (indata->svfname,"a");
        fprintf(outptr,"\n\n\n");
       }
      else outptr = fopen (indata->svfname,"w");
     };
   };

// assign input data from Window Input Structure
  year = indata->year;
  timezone = indata->tzone;
  deltdut = indata->del_tdut / 86400.0;  // difference TDT - UT in days
  indata->eclbuf.lastyear = year;
  indata->eclbuf.lasttz = timezone;
  indata->eclbuf.lastdlt = indata->del_tdut;

// initializing counters
  indata->eclbuf.numecl = 0;
  mp0 = 0;
  mp25 = 0;
  mp5 = 0;
  mp75 = 0;

  yr = year - 0.2;
  yx = year + 1.2;
  day = 1;
  month = 1;
  hour = 0;
  minmjd = mjd(day, month, year, hour);
  day = 31;
  month = 12;
  hour = 24.0;
  maxmjd = mjd(day, month, year, hour);

  /* As the function phmjd finds the respective phases of the Moon
     only to within a few weeks of the time given by the input
     parameter (year.faction) the following loop starts well ahead of
     the beginning of the year in question and ends well after.
     The time step of two weeks utilized to increment the year.fraction
     parameter assures that we catch all the different lunations */

  while (yr < yx)
   {                      // calculate and check the phases 
    ckphase(minmjd,maxmjd,yr,timezone,deltdut, mp0, p0, 0.0, indata);
    ckphase(minmjd,maxmjd,yr,timezone,deltdut, mp25, p25, 0.25, indata);
    ckphase(minmjd,maxmjd,yr,timezone,deltdut, mp5, p5, 0.5, indata);
    ckphase(minmjd,maxmjd,yr,timezone,deltdut, mp75, p75, 0.75, indata);
    yr+=0.02;  // incease by 1/50th of a year (about two weeks) 
   };
  putphase (year,timezone,maxmjd,p0,p25,p5,p75,mp0,mp25,mp5,mp75,outdata);
  solstice (year, timezone, deltdut, outdata); // get solstices and equinoxes
  putecl (year, indata, outdata);  // output eclipse dates

  if (outptr != NULL)
   {
    fclose(outptr);
    outptr = NULL;
   }
}

/*------------------------- caldisp ---------------------------------*/

void caldisp(WinpData* indata, WoutData* outdata)
{
 // display the calendar of the respective year
 //	 and output it in outdata

  char outbuf[128];
  char itxt[11];
  char wkday[7][3] = {"MO","TU","WE","TH","FR","SA","SU"};
  double jd, h;
  int dy[3][42];
  int d, j, m, yr, k, mold, mcur, ix, t, q;

  sprintf(outbuf,"                    Calendar for %4i", indata->year);
  PutNxtLnT (outdata,outbuf);
  sprintf(outbuf,"                    =================");
  PutNxtLnT (outdata, outbuf);

  mold = 1;  // last month
  mcur = 1;  // currently displayed month

  for (q=0; q<4; q++)
   {
    for (t=0; t<3; t++)
     for (j=0; j<42; j++) dy[t][j] = 0;
      for (t=0; t<3; t++)
       {
        jd = mjd(1,mold,indata->year,0.0);
        d = 1;
        j = int(jd);
        j += 2;
        if (j < 0)
         {
          j = -1 - j;
          j %= 7;
          j = 6 - j;
         }
        else j %= 7;  // first weekday of the month (0 = Monday, ... 6 = Sunday)

        do
         {
          dy[t][j] = d;
          jd += 1.0;
          j++;
          caldat(jd, d, m, yr, h);
         } while (m == mold);

        mold = m;
       };

      outbuf[0] = '\0';
      for (t=0; t<3; t++)
       {
        GetMonth (mcur, itxt);
        strcat(outbuf,"           ");
        strcat(outbuf,itxt);
        if (t<2) strcat(outbuf,"        ");
        mcur++;
       }
      PutSpace(outdata);
      PutNxtLnT (outdata, outbuf);

      for (j=0; j<7; j++)
       {
        outbuf[0] = '\0';
        for (t=0; t<3; t++)
         {
          strncpy(itxt,wkday[j],2);
          itxt[2] = '\0';
          strcat(outbuf,itxt);
          for (k=0; k<6; k++)
           {
            ix = j + 7*k;
            if (dy[t][ix] != 0) sprintf(itxt,"%3i",dy[t][ix]);
            else strcpy(itxt,"   ");
            strcat(outbuf,itxt);
           }
          strcat(outbuf,"    ");
         };
        PutNxtLnT (outdata, outbuf);
       };
   }
 }

 /**********************************************************************/
/*------------------------- GetEclTxt ---------------------------------*/

void GetEclTxt (WinpData* indata, int j, char* jtxt)
 {
  // get the data / kind of eclipse info for the j-th eclipse
  // and place it into jtxt

  int p;
  char dts[13];

  sprintf(jtxt, "%2i : ", (j+1));
  dtmstr(indata->eclbuf.eclmjd[j],dts);
  dts[6] = '\0';
  strcat (jtxt,dts);

  p = indata->eclbuf.phase[j];
  switch (p)
   {
    case 1: strcat(jtxt," Par.Sun");
            break;
    case 2:
    case 4: strcat(jtxt," Ann.Sun");
            break;
    case 3:
    case 5: strcat(jtxt," Tot.Sun");
            break;
    case 6: strcat(jtxt," Ann/Tot.");
            break;

    case -1:
    case -2: strcat(jtxt," Pen.Moon");
             break;
    case -3: strcat(jtxt," Par.Moon");
             break;
    case -4: strcat(jtxt," Tot.Moon");
             break;
   };
 }

void PutEclLst (WinpData* indata, WoutData* outdata)
 {
  // Prepare list of Eclipses for display on screen 
  char outbuf[81];
  int j;

  PutSpace(outdata,2);
  sprintf(outbuf, "         Eclipses for %i ", indata->year);  
  PutNxtLn (outdata, outbuf);
  PutSpace(outdata,2);

  for (j=0; j<indata->eclbuf.numecl; j++)
   {
    GetEclTxt (indata, j, outbuf);
    PutNxtLnT (outdata, outbuf);
    PutSpace(outdata);
   };  

  PutSpace(outdata,2);
  strcpy(outbuf, "     Click at respective line or key number to get eclipse details.");
  PutNxtLn (outdata, outbuf);
  PutSpace(outdata,2);
 }        
 
/*------------------------- getlocmag ---------------------------------*/

double getlocmag(double jd, double ep2, double phi, double lamda,
                 double height, Vec3 rs, Vec3 rm, int& totflg)
 {
  /* get magnitude of solar eclipse at local position.
     jd : MJD (UT) of event
     ep2 : correction for apparent sidereal time (in sec)
     phi, lamda : latitude and longitude of observer in radians
     height : height of observer in m
     rs, rm : geocentric position vector of the sun and the moon
     totflg : 1 if total or annular, 0 otherwise

     RETURN: The magnitude of the eclipse (0 if no eclipse)
  */
  const double ds = 218.245445;  // diameter of Sun in Earth radii
  const double dm = 0.544986;   // diameter of Moon in Earth radii
  Vec3  gm, gs, s;
  double dsun, dmoon, asep;
  double elev;

  gm = GeoPos(jd, ep2, phi, lamda, height);
  gs = rs - gm;
  gm = rm - gm;

  // correct for refraction
  s = EquHor(jd, ep2, phi, lamda, gs);
  s = carpol(s);
  elev = s[2];

  if (elev > -3.5e-2)   // cutoff of -2 deg for calculating refraction
   {
    elev = Refract(elev);
    s[2] = s[2] + elev;
    s = polcar (s);
    gs = HorEqu(jd, ep2, phi, lamda, s);
   };

  s = EquHor(jd, ep2, phi, lamda, gm);
  s = carpol(s);
  elev = s[2];
  if (elev > -3.5e-2)   // cutoff of -2 deg for calculating refraction
   {
    elev = Refract(elev);
    s[2] = s[2] + elev;
    s = polcar (s);
    gm = HorEqu(jd, ep2, phi, lamda, s);
   };

  dsun = atan(0.5*ds/abs(gs)); // apparent radius of the Sun
  dmoon = atan(0.5*dm/abs(gm)); // apparent radius of the Moon

  gs = vnorm(gs);
  gm = vnorm(gm);
  asep = fabs(dot(gs, gm));
  if (asep > 1.0) asep = 1.0;
  asep = acos(asep);    // apparent distance between Sun and Moon

  totflg = 0;
  if ((dsun+dmoon) > asep)  // we have an eclipse
   {
    if (fabs(dsun-dmoon) > asep) totflg = 1;
    asep = fabs(dsun + dmoon - asep) / (2.0 * dsun);
   }
  else asep = 0;

  return asep;
 }

/*------------------------- eclstart ---------------------------------*/

int eclstart(WinpData* indata, int j, double *spt, double *ept, int *spp,
                               int p, char *otxt)
 {
  /* get start and end times of the various phases of the eclipse
     j = index of the eclipse
     spt[i] = start time in MJD for phase i
     ept[i] = end time in MJD for phase i
     spp[i] = kind of phase i
      p = phase

      RETURN: the number of different phases of this eclipse.
              Also set the global eclbuf.jdstart and eclbuf.jdstop
              to the (global jd times) of eclipse start and end.

     The step width will be 1 min
   */
  int nump, eflg, pcur, pold, maxp, i, npflg;
  double magecl;  // local magnitude of eclipse at jd
  double jd, step, jdf, d1, d2, elrise;
  double azim, elev, dist, phi, lamda;
  char dts[13];
  char outb[127];
  Vec3 gx;
  Eclipse eclp;

  nump = 0;
  maxp = 0;
  eflg = 0;  // end flag
  pold = 0;
  elev = 0;
  indata->eclbuf.lccnt = 0;
  indata->eclbuf.maxps = -1.0;
  indata->eclbuf.maxelv = -1.0;
  phi = indata->geolat * M_PI / 180.0;
  lamda = indata->geolong * M_PI / 180.0;
  jd = indata->eclbuf.eclmjd[j] - 0.5;  // start 12 hours before maximum
  jd = jd - indata->tzone/24.0;      // use U.T.
  jdf = jd + 1.5;  // emergency stop if something goes wrong with the loop
  step = 1.0/(24.0*60.0);  // stepwidth 1 minute
  indata->eclbuf.ltotb = jd;
  indata->eclbuf.ltote = jd - 1.0;

  do
   {
    if(p < 0)
     {
      pcur = eclp.lunar (jd, indata->del_tdut);
      gx = eclp.GetRMoon();
      magecl = 1.0;  // just set it 1. It's not needed for lunar eclipses.
      elrise = -9.89e-3;  // allow for refraction during rise/set
     }
    else
     {
      pcur = eclp.solar (jd, indata->del_tdut,d1, d2);
      gx = eclp.GetRSun();
      magecl = 0;
      elrise = -1.45e-2;  // allow for refraction during rise/set
     };

    // check whether body is visible from local position.
    if (pcur > 0)
     {
      AppPos (jd, eclp.GetEp2(), phi, lamda, indata->geoheight, 1, gx,
                                 azim, elev, dist);
      if ((p > 0) && (elev >= elrise))
       {
        magecl = getlocmag(jd, eclp.GetEp2(), phi, lamda, indata->geoheight,
                               eclp.GetRSun(),eclp.GetRMoon(), i);
        if (magecl > indata->eclbuf.maxps)
         {
          indata->eclbuf.maxps = magecl;
          indata->eclbuf.maxelv = elev;
          indata->eclbuf.jdmaxps = jd + indata->tzone/24.0;
         }
       }
     }

    if ((indata->eclbuf.lccnt == 0) && (pcur != 0))
     {
      if ((elev >= elrise) && (magecl > 0))
       {
        indata->eclbuf.lcb1 = jd + indata->tzone/24.0;
        indata->eclbuf.lccnt = 1;
       }
     }
    if (indata->eclbuf.lccnt == 1)
     {
      if ((elev < elrise) || (pcur == 0) || (magecl == 0))
       {
        indata->eclbuf.lce1 = jd + indata->tzone/24.0;
        indata->eclbuf.lccnt = 2;
       }
     }
    if ((indata->eclbuf.lccnt == 2) && (pcur != 0))
     {
      if ((elev >= elrise) && (magecl > 0))
       {
        indata->eclbuf.lcb2 = jd + indata->tzone/24.0;
        indata->eclbuf.lccnt = 3;
       }
     }
    if (indata->eclbuf.lccnt == 3)
     {
      if ((elev < elrise) || (pcur == 0) || (magecl == 0))
       {
        indata->eclbuf.lce2 = jd + indata->tzone/24.0;
        indata->eclbuf.lccnt = 4;
       }
     }

    // now check the start and stop times for the phases
    if (pcur > pold)
     {
      if(nump < 4)
       {
        spt[nump] = jd + indata->tzone/24.0;  // back to local time
        spp[nump] = pcur;
        nump++;
        maxp++;
        pold = pcur;
       }
     }
    else if (pcur < pold)
     {
      npflg = 1;
      if (nump > 1)
       {
        if (pcur > spp[nump-2]) npflg = 0;
       }
      if (npflg)
       {
        nump--;
        if (nump >= 0) ept[nump] = jd + indata->tzone/24.0;
        if (nump > 1)
         {
          if (pcur < spp[nump-1])
           {
            nump--;
            ept[nump] = jd + indata->tzone/24.0;
           }
         }
        if (nump <= 0) eflg = 1;
       }
      pold = pcur;
     };

    jd += step;
    if (jd > jdf) eflg = 1;

   } while (eflg != 1);

  //  put the respective info into textfile otxt
  if(maxp > 0)
   {
    for (i=0; i<maxp; i++)
     {
      if (p < 0)
       {
        switch (spp[i])
         {
          case 1: strcpy(outb,"penumbral   ");
                  break;
          case 2: strcpy(outb,"tot.penumb  ");
                  break;
          case 3: strcpy(outb,"partial     ");
                  break;
          case 4: strcpy(outb,"totality    ");
                  break;
         }
       }
      else
       {
        switch (spp[i])
         {
          case 1: strcpy(outb,"partial     ");
                  break;
          case 2: strcpy(outb,"n-centr.Ann ");
                  break;
          case 3: strcpy(outb,"n-centr.Tot ");
                  break;
          case 4: strcpy(outb,"annularity  ");
                  break;
          case 5: strcpy(outb,"totality    ");
                  break;
         }
       }

      if (strlen(otxt) <= 200)  // just make sure the buffer doesn't overflow
       {
        strcat(otxt,"\n ");
        strcat(otxt, outb);
        strcat(otxt,"\tbegins ");
        dtmstr(spt[i],dts);
        dts[12] = '\0';
        strcat(otxt,dts);
        strcat(otxt, "\tends ");
        dtmstr(ept[i],dts);
        dts[12] = '\0';
        strcat(otxt,dts);
       }
     }
   }
  if (maxp > 0)
   {
    if (indata->eclbuf.lccnt == 1) indata->eclbuf.lce1 = ept[0];
    if (indata->eclbuf.lccnt == 3) indata->eclbuf.lce2 = ept[0];
   }

  if (indata->eclbuf.maxps > 0.8)  // check for local central phase
   {
    jd = indata->eclbuf.jdmaxps - indata->tzone/24.0 - 16.0/(24.0*60.0);
    indata->eclbuf.ltote = jd + 32.0/(24.0*60.0);
    eflg = 0;
    do
     {
      jd += 1.0/86400.0;  // go in seconds steps
      eclp.solar (jd, indata->del_tdut,d1, d2);
      gx = eclp.GetRSun();
      AppPos (jd, eclp.GetEp2(), phi, lamda, indata->geoheight, 1, gx,
                                      azim, elev, dist);
      if (elev >= elrise)
       {
        magecl = getlocmag(jd, eclp.GetEp2(), phi, lamda, indata->geoheight,
                               eclp.GetRSun(),eclp.GetRMoon(), i);
        if (magecl > indata->eclbuf.maxps)
         {
          indata->eclbuf.maxelv = elev;
          indata->eclbuf.maxps = magecl;
          indata->eclbuf.jdmaxps = jd + indata->tzone/24.0;
         }
        if (i > 0)
         {
          indata->eclbuf.ltotb = jd;
          eflg = 1;
         }
       }
      if (jd > indata->eclbuf.ltote) eflg = 1;

     } while (!eflg);

    if (jd < indata->eclbuf.ltote)
     {
      eflg = 0;
      do
       {
        jd += 1.0/86400.0;  // go in seconds steps
        eclp.solar (jd, indata->del_tdut,d1, d2);
        gx = eclp.GetRSun();
        AppPos (jd, eclp.GetEp2(), phi, lamda, indata->geoheight, 1, gx,
                                        azim, elev, dist);
        if (elev < elrise) eflg = 1;
        magecl = getlocmag(jd, eclp.GetEp2(), phi, lamda, indata->geoheight,
                               eclp.GetRSun(),eclp.GetRMoon(), i);
        if (magecl > indata->eclbuf.maxps)
         {
          indata->eclbuf.maxelv = elev;
          indata->eclbuf.maxps = magecl;
          indata->eclbuf.jdmaxps = jd + indata->tzone/24.0;
         }
        if (i == 0) eflg = 1;
        if (jd > indata->eclbuf.ltote) eflg = 1;

       } while (!eflg);
      indata->eclbuf.ltote = jd + indata->tzone/24.0;
      indata->eclbuf.ltotb += indata->tzone/24.0;
     }
    else indata->eclbuf.ltote = indata->eclbuf.ltotb - 1.0;
   }

  indata->eclbuf.jdstart = spt[0];
  indata->eclbuf.jdstop = ept[0];

  return maxp;
 }

/*------------------------- hdfdms ---------------------------------*/

double hdfdms (double h)
 {
  // convert degrees from d.fff -> d.mm where .mm are the minutes

  int mm, deg;
  double hh, sec;

  hh = fabs(h);
  dms (hh,deg,mm,sec);
  if (sec>30.0) {mm++;};
  if (mm>59)
   {
    deg++;
    mm=0;
   };
  hh = double(deg);
  sec = double(mm)/100.0;
  hh += sec;
  if (h < 0) hh = -hh;

  return hh;
 }

/*------------------------- DegFDms ---------------------------------*/

double DegFDms (double h)
 {
  /* convert degrees from d.fff -> d.mmsss where .mm are the minutes
     and sss are seconds (and fractions of seconds).
  */
  int mm, deg;
  double hh, t, sec;

  hh = fabs(h);
  dms (hh,deg,mm,sec);
  if (sec>=59.5)
   {
    mm++;
    sec = 0;
   };
  if (mm>59)
   {
    deg++;
    mm=0;
   };
  hh = double(deg);
  t = double(mm)/100.0;
  hh += t;
  t = sec/10000.0;
  hh += t;
  if (h < 0) hh = -hh;

  return hh;
 }

/*------------------------- DmsDegF ---------------------------------*/

double DmsDegF (double h)
 {
  /* convert degrees from d.mmsss -> d.fff where .mm are the minutes
     and sss are seconds (and fractions of seconds).
     Returned is the angle in decimal degrees.
  */
  int mm, deg;
  double hh, t, sec, dlt;

  hh = fabs(h);
  dlt = hh*3.33e-16;
  hh += dlt;
  deg = int(hh);
  t = fmod(hh,1.0)*100.0;
  mm = int(t + 0.1e-12);
  sec = fmod(hh*100.0,1.0)*100.0;
  hh = ddd(deg,mm,sec);
  if (h < 0) hh = -hh;

  return hh;
 }

/*----------------------------- SatEcl ---------------------------------*/

void SatEcl(double jd, double tdut, double height, bool geoid, int &et1, double &lat1,
            double &lng1, int &et2, double &lat2, double &lng2)
{
 /* Get one set of central eclipse coordinates for MJD jd (given in UT) 
    tdut : delta TT - UT
    height : height above Earth Geoid in km
    geoid : true if surface flattening
    et1 : 0 if no central eclipse, 1 if total, 2 if annular  (front surface)
    lat1, lng1 : coordinates of central eclipse at time jd in decimal degrees
    et2, lat2, lng2 as above except for the back surface
 */
  double ae = 23454.77992; // 149597870.0/6378.14 =  1AE -> Earth Radii
  double flat = 0.996633;  // flatting of the Earth
  double ds = 218.245445;  // diameter of Sun in Earth radii
  double dm = 0.544986;   // diameter of Moon in Earth radii
  double t;      // time in Julian Centuries
  double ep2;    // correction for Apparent Sideral Time
  double d_umbra; // diameter of umbra in Earth Radii
  double d_penumbra; // diameter of penumbra in Earth Radii
  double s0, s, dlt, r2, r0,lamda, phi, zs;
  int phase;
  Mat3 mx;
  Sun200 sun;
  Moon200 moon;
  Vec3 rs, rm;   // position of the Sun and the Moon
  Vec3 eshadow;  // unit vector in direction of shadow
  Vec3 rint;     // intersection shadow axis - Earth surface
  Vec3 ve;

  t = julcent (jd) + tdut / 3.15576e9;  // =(86400.0 * 36525.0);
  rs = sun.position(t);
  rm = moon.position(t);
  rs = eclequ(t,rs);
  rm = eclequ(t,rm);

  // correct Moon coordinates for center of figure
  mx = zrot(-2.4240684e-6); // +0.5" in longitude
  rm = mxvct(mx,rm);
  mx = yrot(-1.2120342e-6); // -0.25" in latitude
  rm = mxvct(mx,rm);
  mx = nutmat(t,ep2);   // nutation
  rs = mxvct(mx,rs);    // apparent coordinates
  rs = aberrat(t,rs);
  rs *= ae;
  rm = mxvct(mx,rm);

  // correct for surface height
  s = height;
  if (s < 0) s = 0;
  zs = 6378.14 / (6378.14 + s);
  ds *= zs;
  dm *= zs;
  rs *= zs;
  rm *= zs;

  if (geoid)
   {
    rs[2]/=flat;  // adjust for flatting of the Earth
    rm[2]/=flat;
   }; 
  lamda = 0;
  phi = 0;
  et1 = 0;
  et2 = 0;

  // intersect shadow axis with Earth
  eshadow = rm - rs;
  eshadow = vnorm(eshadow);    // direction vector of shadow

  s0 = - dot(rm, eshadow);   // distance Moon - fundamental plane
  r2 = dot (rm,rm);
  dlt = s0*s0 + 1.0 - r2;
  r0 = 1.0 - dlt;
  if (r0 > 0) r0 = sqrt (r0);
  else r0 = 0;      // distance center of Earth - shadow axis
  r2 = abs(rs - rm);
  d_umbra = (ds - dm) * s0 / r2 - dm;// diameter of umbra at fundamental plane
  d_penumbra = (ds + dm) * s0 / r2 + dm;

  // get phase of eclipse
  if (r0 < 1.0)
   {
    if (dlt > 0) dlt = sqrt(dlt);
    else dlt = 0;

    // first the front (daylight) side
    s = s0 - dlt;  // distance Moon - fundamental plane
    d_umbra = (ds - dm) * s / r2 - dm; // diameter of umbra at surface
    rint = rm + s * eshadow;
    rint[2] *= flat;    // vector to intersection
    ve = carpol(rint);
    lamda = ve[1] - lsidtim(jd,0,ep2)*0.261799387799; // geographic coordinates
    if (lamda > M_PI) lamda -= 2.0*M_PI;
    if (lamda < (-M_PI)) lamda += 2.0*M_PI;
    phi = sqrt(rint[0]*rint[0] + rint[1]*rint[1])*0.993305615;
    phi = atan2(rint[2],phi);
    lat1 = phi / degrad;
    lng1 = lamda / degrad;
    if (d_umbra > 0) phase = 4;  // central annular eclipse
    else phase = 5;              // central total eclipse
    if (phase == 4) et1 = 2;
    if (phase == 5) et1 = 1;

    // now the back (night) side
    if (r0 > zs)  // make sure the Earth is not in the way
     {
      s = s0 + dlt;  // distance Moon - fundamental plane
      d_umbra = (ds - dm) * s / r2 - dm; // diameter of umbra at surface
      rint = rm + s * eshadow;
      rint[2] *= flat;    // vector to intersection
      ve = carpol(rint);
      lamda = ve[1] - lsidtim(jd,0,ep2)*0.261799387799; // geographic coordinates
      if (lamda > M_PI) lamda -= 2.0*M_PI;
      if (lamda < (-M_PI)) lamda += 2.0*M_PI;
      phi = sqrt(rint[0]*rint[0] + rint[1]*rint[1])*0.993305615;
      phi = atan2(rint[2],phi);
      lat2 = phi / degrad;
      lng2 = lamda / degrad;
      if (d_umbra > 0) phase = 4;  // central annular eclipse
      else phase = 5;              // central total eclipse
      if (phase == 4) et2 = 2;
      if (phase == 5) et2 = 1;
     };
   };
}

/*------------------------- EclCentral ---------------------------------*/

void EclCentral(int first, int p, double& jd, double jdf, WinpData* indata,
                                  char *dcmp, char *otxt)
 {
  /* output one line of central eclipse data for MJD jd (given in UT)
     first = 1 for first call
     p = phase (= 4 annular, = 5 total, = 6 annular/total
     jdf = MJD of stop time
     dcmp = buffer to compare last date outputted
     otxt = text buffer
  */
  double phi, lamda, djd, width;
  double azim, elev, dist;
  static int lcnt;  // line count
  static double mpelev;  // elevation for the maximum phase
  int kp, sln;
  char dts[13];
  char ndts[91];
  Vec3 gx;
  Eclipse eclp;

  kp = eclp.solar(jd, indata->del_tdut, phi, lamda);
  indata->eclbuf.cphs = 0;

  if (first == 1)   // find the first occurence compliant with the step width
   {
    mpelev = -10.0;
    while ((kp < 4) && (jd < jdf))
     {
      jd += double(indata->eclbuf.cstep) / (24.0*60.0);
      kp = eclp.solar(jd, indata->del_tdut, phi, lamda);
     };
    lcnt = 0;
   }

  if (kp > 3)   // central eclipse
   {
    if ((lcnt % 10) == 0)
             strcpy(otxt,"\n\nhh:mm\t Lat.\t Long.\t\tsec\tkm \telev\n");
    else strcpy(otxt,"\n");
    lcnt++;

    djd = jd + indata->tzone/24.0;
    dtmstr(djd,dts);
    dts[12] = '\0';
    if (strncmp(dts,dcmp,6) != 0)
     {
      strncpy(dcmp,dts,6);
      strcat(otxt,dcmp);
      strcat(otxt,"\n");
     };

    indata->eclbuf.cphs = kp;
    if (p != 6) strcat(otxt,"  ");
    else
     {    // in case of a annular/total mark 'a' or 't'
      if (kp == 4) strcat(otxt,"a ");
      else strcat(otxt,"t ");
     }

    gx = eclp.GetRSun();
    AppPos (jd, 0, phi, lamda, 0, 1, gx, azim, elev, dist);
    width = 180.0/M_PI;
    phi *= width;
    lamda *= width;
    elev *= width;
    if (lamda > 180.0) lamda -= 360.0;
    if (indata->dmsflg == 1)
     {
      phi = hdfdms(phi);
      lamda = hdfdms(lamda);
     }
    indata->eclbuf.clat = phi;
    indata->eclbuf.clng = lamda;
    if (elev > mpelev)
     {
      mpelev = elev;
      indata->eclbuf.cmxlat = phi;
      indata->eclbuf.cmxlng = lamda;
     };

    sln = strlen(otxt);
    for (kp=0; kp < 5; kp++) otxt[sln+kp] = dts[7+kp];
    otxt[sln+5] = '\0';
    djd = eclp.duration(jd, indata->del_tdut, width);
    sprintf(ndts,"\t%7.2f\t%8.2f\t\t%3.0f\t%4.0f\t%3.0f",
                             phi,lamda,djd,width,elev);
    decdot(ndts);
    strcat(otxt,ndts);
   }
 }

/*------------------------- GetMaxPos ---------------------------------*/

void GetMaxPos(WinpData* indata)
 {
  /* Get the geographic position of the maximum eclipse in case
     of a partial or non-central eclipse and place the result in
     indata->eclbuf.cmxlat and indata->eclbuf.cmxlng
     NOTE : This function will only return meaningful values
            in case of a partial or non-central eclipse!
  */

  int j;
  double t;
  Eclipse eclp;

  j = indata->eclbuf.eclselect-1;

  if (indata->eclbuf.phase[j] < 1)
   {
    indata->eclbuf.cmxlat = 0;
    indata->eclbuf.cmxlng = 0;
   }
  else
   {
    t = indata->eclbuf.eclmjd[j] - indata->tzone/24.0;
    eclp.maxpos(t,indata->del_tdut,indata->eclbuf.cmxlat,indata->eclbuf.cmxlng);
    indata->eclbuf.cmxlat /= degrad;
    indata->eclbuf.cmxlng /= degrad;
   };
 }

/*------------------------- InitBound ---------------------------------*/

void InitBound(WinpData* indata, Vec3& ubm, Vec3& ube, Vec3& udm, Vec3& ude,
                                 Vec3& lbe, Vec3& lde)
 {
  /* Initialize the calculation for the northern and southern boundaries
     of solar eclipses.
     The function EclStart must have been called first to enable InitBound
     to get the right eclipse.
     The calculation is done in Earth radii.

     OUTPUT:
            ubm : Penumbra base vector
            ube : Shadod base vector for upper boundary
            udm : Penumbra base delta vector
            ude : Shadow delta vector for upper boundary
            lbe : Shadod base vector for lower boundary
            lde : Shadow delta vector for lower boundary
  */

//  const double dm = 0.272493;   // radius of Moon in Earth radii
  double dpn1, dpn2, pan1, pan2, s0;
  Vec3 shmv, u2m, u2e;
  Vec3 ax1, ax2;
  Eclipse eclp;

  // beginning of the eclipse
  eclp.penumd (indata->eclbuf.jdstart, indata->del_tdut, ubm, ube, dpn1, pan1);

  // end of the eclipse
  eclp.penumd (indata->eclbuf.jdstop, indata->del_tdut, u2m, u2e, dpn2, pan2);

  dpn1 *= 0.5;
  dpn2 *= 0.5;

  // get apex of penumbra cone
  pan1 = tan(pan1);
  if (pan1 != 0) dpn1 = dpn1 / pan1;
  else dpn1 = 1.2 * abs(ubm);  // avoid a crash
  pan2 = tan(pan2);
  if (pan2 != 0) dpn2 = dpn2 / pan2;
  else dpn2 = dpn1;  // avoid a crash

  // get vector perpendicular to movement of shadow
  s0 = - dot(ubm, ube);
  ax1 = ubm + s0 * ube;
  ubm = ubm + (s0 - dpn1) * ube;
  s0 = - dot(u2m, u2e);
  ax2 = u2m + s0 * u2e;
  u2m = u2m + (s0 - dpn2) * u2e;
  shmv = ax1 - ax2;
  ax2 = ax1 * ax2;
  shmv = shmv * ax2;
  shmv = vnorm(shmv);

  // now get the delta vectors
  udm = u2m - ubm;
  lbe = ube - pan1 * shmv;
  ube = ube + pan1 * shmv;
  ude = u2e + pan2 * shmv;
  lde = u2e - pan2 * shmv;
  ube = vnorm(ube);
  lbe = vnorm(lbe);
  ude = vnorm(ude);
  lde = vnorm(lde);
  lde = lde - lbe;
  ude = ude - ube;

  // scale the delta vectors
  dpn1 = indata->eclbuf.jdstop - indata->eclbuf.jdstart;
  if (dpn1 == 0) dpn1 = 1.0;
  else dpn1 = 1.0 / dpn1;
  udm *= dpn1;
  ude *= dpn1;
  lde *= dpn1;
 }

/*------------------------- GNSBound ---------------------------------*/

void GNSBound(WinpData* indata, double t, Vec3& ubm, Vec3& ube, Vec3& udm,
                                          Vec3& ude, double& lat, double& lng)
 {
  /* Get the geographic coordinates of the northern or southern boundaries
     at time t.
     INPUT: see InitBound
     OUTPUT:
            lat, lng : latitude and longitude of penumbra boundary.
            If lat > 90 no northern boundary exists.
   */
  const double flat = 0.996633;  // flatting of the Earth

  double s0, s, r0, r2, dlt;
  Vec3 vrm, vre;

  // get shadow vector at time t
  vrm = ubm + (t - indata->eclbuf.jdstart) * udm;
  vre = ube + (t - indata->eclbuf.jdstart) * ude;
  vre = vnorm (vre);  // direction of penumbra boundary

  s0 = - dot(vrm, vre);   // distance Moon - fundamental plane
  r2 = dot (vrm,vrm);
  dlt = s0*s0 + 1.0 - r2;
  r0 = 1.0 - dlt;

  if (r0 > 0) r0 = sqrt (r0);
  else r0 = 0;      // distance center of Earth - shadow axis

  // calculate the coordinates if there is an intersecton
  if (r0 < 1.0)  // there should be an intersection
   {
    if (dlt > 0) dlt = sqrt(dlt);
    else dlt = 0;
    s = s0 - dlt;  // distance Moon - fundamental plane
    vrm = vrm + s * vre;
    vrm = vnorm(vrm);
    vrm[2] *= flat;    // vector to intersection
    vre = carpol(vrm);
    lng = vre[1] - lsidtim(t,0,0)*0.261799387799; // geographic coordinates
    if (lng > M_PI) lng -= 2.0*M_PI;
    if (lng < (-M_PI)) lng += 2.0*M_PI;
    lat = sqrt(vrm[0]*vrm[0] + vrm[1]*vrm[1])*0.993305615;
    lat = atan23(vrm[2],lat);
    lat /= degrad;
    lng /= degrad;
   }
  else
   {
    lat = 100.0;
    lng = 0;
   };

 }

/*------------------------- InRSBound ---------------------------------*/

void InRSBound(WinpData* indata, Vec3& ubm, Vec3& ube, Vec3& udm, Vec3& ude,
                                 double& dpb, double& dpd)
 {
  /* Initialize the calculation for the sunris and sunset boundaries
     of solar eclipses.
     The function EclStart must have been called first to enable InRSBound
     to get the right eclipse.
     The calculation is done in Earth radii.

     OUTPUT:
            ubm : Moon base vector
            ube : Shadod base vector
            udm : Moon delta vector
            ude : Shadow delta vector for
            dpb : Base value for diameter of penumbra
            dpd : delta value for diameter of penumbra
  */

  double pan;
  Vec3 u2m, u2e;
  Eclipse eclp;

  // beginning of the eclipse
  eclp.penumd (indata->eclbuf.jdstart, indata->del_tdut, ubm, ube, dpb, pan);

  // end of the eclipse
  eclp.penumd (indata->eclbuf.jdstop, indata->del_tdut, u2m, u2e, dpd, pan);

  // get the delta vectors
  udm = u2m - ubm;
  ude = u2e - ube;
  dpd = dpd - dpb;

  // scale the delta vectors

  pan = indata->eclbuf.jdstop - indata->eclbuf.jdstart;
  if (pan == 0) pan = 1.0;
  else pan = 1.0 / pan;
  udm *= pan;
  ude *= pan;
  dpd *= pan;
 }

/*------------------------- GRSBound ---------------------------------*/

int iscrs(double vrc0, double vrc1, double dpn,
                       double& vrx0, double& vrx1, double& vrx20, double& vrx21)
 {
  /* calculate intersection vectors of the penumbral cone with the
     Earth in the fundamental plane to find the locations of rise
     and set.

     INPUT : components of vectors vrc (pointing to the center of the
             penumbra), vre (unit vector perpendicular to the
             fundamental plane).
             dpn : half diameter of penumbra in Earth radii

     OUTPUT: components of the intersection vectors vrx and vrx2 if
             successful.

     RETURN: 1 if intersection successful, 0 otherwise.
  */
  int rtn;
  double a, b, c, f1, f2, f3, f4;

  rtn = 1;
  f1 = vrc0*vrc0;

  if (f1 < 1.0e-60) rtn = 0;
  else
   {
    f2 = 1.0 - dpn*dpn + f1 + vrc1*vrc1;
    f3 = f2 / (2.0*vrc0);
    f4 = vrc1 / vrc0;
    a = 1.0 + vrc1*vrc1 / f1;
    b = f2*vrc1 / f1;
    c = f2*f2 / (4.0*f1) - 1.0;
    if (fabs(a) < 1.0e-60) rtn = 0;
    else
     {
      vrx21 = -0.5 * b / a; ;
      vrx0 = vrx21*vrx21 - c / a;
      if (vrx0 < 0) rtn = 0;
      else
       {
        vrx0 = sqrt(vrx0);
        vrx1 = vrx21 + vrx0;
        vrx21 = vrx21 - vrx0;
        vrx0 = f3 + f4*vrx1;
        vrx20 = f3 + f4*vrx21;
        vrx1 = -vrx1;
        vrx21 = -vrx21;
       };
     };
   };

  return rtn;
 }

void GRSBound(WinpData* indata, double t, Vec3& ubm, Vec3& ube, Vec3& udm,
                                Vec3& ude, double dpb, double dpd,
                                double& lat1, double& lng1, double& lat2, double& lng2)
 {
  /* Get the geographic coordinates of the boundaries for rise/set
     at time t.

     INPUT: t : MJD of time
            ubm, ube, udm, ude, dpb, dpd see InRSBound.

     OUTPUT:
            lat1, lng1; lat2, lng2  : latitude and longitude of first
                                      and second point ofboundary.
                                      If lat > 90 no respective point exists.
  */
  const double flat = 0.996633;  // flatting of the Earth

  double s0, r0, dpn;
  Vec3 vrm, vre, vrc, vrx, vrx2;
  Mat3 m1, m2;

  // get position of penumbra and shadow cone
  vrm = ubm + (t - indata->eclbuf.jdstart) * udm;
  vre = ube + (t - indata->eclbuf.jdstart) * ude;
  vre = vnorm (vre);  // direction of penumbra boundary
  dpn = dpb + (t - indata->eclbuf.jdstart) * dpd;
  dpn *= 0.5;  // half cone diameter
  vrx = carpol(vre);

  m1 = zrot(vrx[1] + M_PI/2.0);
  m2 = xrot(M_PI/2.0 - vrx[2]);
  m1 = m2 * m1;   // rotation from equatorial into fundamental plane
  m2 = mxtrn(m1); // rotation from fundamental plane into equatorial

  // get vector to center of shadow in the fundamental plane
  s0 = - dot(vrm, vre);
  vrc = vrm + s0 * vre;

  r0 = abs(vrc);  // distance between center of Earth and center of shadow
  vrc = mxvct(m1, vrc);

  lng1 = 0;
  lat1 = 100.0;
  lng2 = 0;
  lat2 = 100.0;
  // check whether intersection of Earth and shadow cone are possible
  // in fundamental plane
  if ((r0 > fabs(1.0 - dpn)) && (r0 < fabs(1.0 + dpn)))
   {
    // now find the intersections
    if (iscrs(vrc[0],vrc[1], dpn, vrx[0],vrx[1], vrx2[0],vrx2[1])) lat1 = 0;
    else
     {
      if (iscrs(vrc[1],vrc[0], dpn, vrx[1],vrx[0], vrx2[1],vrx2[0])) lat1 = 0;
     };
   };

  // calculate the coordinates if there is an intersecton
  if (lat1 < 100.0)  // there should be an intersection
   {
    vrx[2] = 0;
    vrx = mxvct(m2, vrx);
    vrx = vnorm(vrx);
    vrx[2] *= flat;    // vector to intersection
    vre = carpol(vrx);
    lng1 = vre[1] - lsidtim(t,0,0)*0.261799387799; // geographic coordinates

    if (lng1 > M_PI) lng1 -= 2.0*M_PI;
    if (lng1 < (-M_PI)) lng1 += 2.0*M_PI;
    lat1 = sqrt(vrx[0]*vrx[0] + vrx[1]*vrx[1])*0.993305615;
    lat1 = atan23(vrx[2],lat1);
    lat1 /= degrad;
    lng1 /= degrad;
   };
  if (lat1 < 100.0)  // intersection #2
   {
    vrx2[2] = 0;
    vrx2 = mxvct(m2, vrx2);
    vrx2 = vnorm(vrx2);
    vrx2[2] *= flat;    // vector to intersection
    vre = carpol(vrx2);
    lng2 = vre[1] - lsidtim(t,0,0)*0.261799387799; // geographic coordinates
    if (lng2 > M_PI) lng2 -= 2.0*M_PI;
    if (lng2 < (-M_PI)) lng2 += 2.0*M_PI;
    lat2 = sqrt(vrx2[0]*vrx2[0] + vrx2[1]*vrx2[1])*0.993305615;
    lat2 = atan23(vrx2[2],lat2);
    lat2 /= degrad;
    lng2 /= degrad;
   };
 }

/*------------------------- EclDetails ---------------------------------*/

void EclDetails(WinpData* indata, char *otxt)
 {
  /* get the details of the eclipse selected in eclbuf.select
     and place the output into otxt
  */
  int j, p, i;
  int dd, mm, yy, deg, mnt;
  double sec, hh;

  static int spp[4], nump;
  static double spt[4], ept[4];
  static double jd, jdf;
  static char dcmp[7];
  char dts[13];
  char outb[127];

  j = indata->eclbuf.eclselect-1;
  p = indata->eclbuf.phase[j];

  if (indata->eclbuf.ecloutbn == 1)
   {
    switch (p)
     {
      case 1: strcpy(otxt,"\t\tPartial Eclipse of the Sun");
              break;
      case 2: strcpy(otxt,"\t\tNon-Central Annular Eclipse of the Sun");
              break;
      case 3: strcpy(otxt,"\t\tNon-Central Total Eclipse of the Sun");
              break;
      case 4: strcpy(otxt,"\t\tAnnular Eclipse of the Sun");
              break;
      case 5: strcpy(otxt, "\t\tTotal Eclipse of the Sun");
              break;
      case 6: strcpy(otxt, "\t\tAnnular/Total Solar Eclipse");
              break;

      case -1:
      case -2: strcpy(otxt, "\t\tPenumbral Eclipse of the Moon");
               break;
      case -3: strcpy(otxt, "\t\tPartial eclipse of the Moon");
               break;
      case -4: strcpy(otxt, "\t\tTotal eclipse of the Moon");
               break;
     };

    sprintf(outb,"\n\nMaximum Eclipse at ");
    strcat(otxt,outb);
    dtmstr(indata->eclbuf.eclmjd[j],dts);
    dts[12] = '\0';
    strcat(otxt,dts);
    if (p < 4)
     {
      sprintf(outb,"   with magnitude:%5.2f",indata->eclbuf.magnitude[j]);
      decdot0(outb);
      strcat(otxt, outb);
     }
   }

  if (indata->eclbuf.ecloutbn == 2)
   {
    strcpy(otxt,"\n");
    nump = eclstart(indata, j, spt, ept, spp, p, otxt);
    if ((p < 4) || (nump < 1)) indata->eclbuf.ecloutbn = 5;
   }

  if (indata->eclbuf.ecloutbn == 3)
   {
    // get start and stop dates for central phase
    jd = spt[nump-1];
    for (i=0; i<nump; i++)
    if ((spt[i] < jd) && (spp[i] > 3)) jd = spt[i]; // start of central phase
    caldat(jd, dd, mm, yy, hh);
    dms (hh,deg,mnt,sec);
    sec = 0;
    i = mnt / indata->eclbuf.cstep;
    mnt = i*indata->eclbuf.cstep;   // cut to proper time step
    hh = ddd (deg, mnt, sec);
    jd = mjd (dd, mm, yy, hh);
    jd = jd - indata->tzone/24.0;      // use UT
    jdf = ept[nump-1];
    for (i=0; i<nump; i++)
     if ((ept[i] > jdf) && (spp[i] > 3)) jdf = ept[i]; // end of central phase
    jdf = jdf - indata->tzone/24.0;      // use UT

    strcpy(dcmp,"******");  // compare the date to know when to print it

    // get first output line for central eclipse
    EclCentral(1, p, jd, jdf, indata, dcmp, otxt);
    jd += double(indata->eclbuf.cstep) / (24.0*60.0);
   }

  if (indata->eclbuf.ecloutbn == 4)   // consecutive central eclipse data
   {
    EclCentral(0, p, jd, jdf, indata, dcmp, otxt);
    jd += double(indata->eclbuf.cstep) / (24.0*60.0);
    if (jd > jdf) indata->eclbuf.ecloutbn = 5;
    else indata->eclbuf.ecloutbn--;
   }

  if (indata->eclbuf.ecloutbn == 6) // local circumstances
   {
    if (indata->locecl) strcpy (otxt, " ");
    else
     {
      strcpy(otxt, "\n\n\nLocal Circumstances for ");
      strncat(otxt, indata->geoname, 30);
     };
    if (indata->dmsflg)  // deg.fff -> DD.MMSSS
     {
      jd = DegFDms(indata->geolat);
      jdf = DegFDms(indata->geolong);
     }
    else
     {
      jd = indata->geolat;
      jdf = indata->geolong;
     }
    sprintf(outb,"\nLat: %g   Long: %g   height: %g m\n\n",
                               jd, jdf, indata->geoheight);
    decdot0(outb);
    strcat(otxt,outb);
    if (p != 0)
     {
      if (indata->eclbuf.lccnt > 0)
       {
        sprintf(outb,"Eclipse visible from ");
        dtmstr(indata->eclbuf.lcb1,dts);
        dts[12] = '\0';
        strcat(outb," ");
        strcat(outb,dts);
        strcat(outb," to ");
        dtmstr(indata->eclbuf.lce1,dts);
        dts[12] = '\0';
        strcat(outb,dts);
        if (indata->eclbuf.lccnt > 2)  // this case almost never happens
         {
          strcat(otxt,outb);
          strcpy(outb,"\n\tand from ");
          dtmstr(indata->eclbuf.lcb2,dts);
          dts[12] = '\0';
          strcat(outb," ");
          strcat(outb,dts);
          strcat(outb," to ");
          dtmstr(indata->eclbuf.lce2,dts);
          dts[12] = '\0';
          strcat(outb,dts);
         }
       }
      else sprintf(outb,"Eclipse not visible");
      strcat(otxt,outb);
     }
   }

  if (indata->eclbuf.ecloutbn == 7) // local solar eclipse magnitude
   {
    if (indata->locecl) strcpy (otxt, " ");
    if ((p > 0) && (indata->eclbuf.lccnt > 0))
     {
      sprintf(otxt,"\nMaximum Eclipse at ");
      dtmstr(indata->eclbuf.jdmaxps,dts);
      dts[12] = '\0';
      strcat(otxt,dts);
      sprintf(outb,"   with magnitude %6.3f", indata->eclbuf.maxps);
      decdot0(outb);
      strcat(otxt,outb);
      sprintf(outb,"   elev:%4.1f", 180.0*indata->eclbuf.maxelv/M_PI);
      decdot0(outb);
      strcat(otxt,outb);
      if (indata->eclbuf.ltotb <= indata->eclbuf.ltote)
       {      // local central eclipse
        if ((p % 2) == 0) strcpy(outb, "\nannularity from");
        else strcpy(outb, "\ntotality from");
        if ((p == 1) || (p == 6))
                   strcpy(outb, "\ntotality/annularity from");
        strcat(otxt,outb);
        caldat(indata->eclbuf.ltotb, dd, mm, yy, hh);
        caldat(indata->eclbuf.ltote, dd, mm, yy, jd);
        hh = DegFDms(hh);
        jd = DegFDms(jd);
        jdf = (indata->eclbuf.ltote - indata->eclbuf.ltotb) * 86400.0;
        sprintf(outb,"%8.4f  to%8.4f   del.t:%3.0f sec",hh,jd, jdf);
        decdot0(outb);
        strcat(otxt,outb);
       }
     }
    indata->eclbuf.ecloutbn = 0;
   }
 }

/*------------------------- EclPltCen ---------------------------------*/

void EclPltCen(WinpData* indata, char *otxt)
 {
  /* get the details of the eclipse selected in eclbuf.select
     and provide the central line coordinates on successive calls
     to indata->eclbuf.clat and .clng for plotting.
     Also place the text output into otxt
  */

  int j, p, i;
  int dd, mm, yy, deg, mnt;
  double sec, hh;
  static int spp[4], nump;
  static double spt[4], ept[4];
  static double jd, jdf;
  static char dcmp[7];

  j = indata->eclbuf.eclselect-1;
  p = indata->eclbuf.phase[j];

  if (indata->eclbuf.ecloutbn == 2)
   {
    strcpy(otxt,"\n");
    nump = eclstart(indata, j, spt, ept, spp, p, otxt);
    if ((p < 4) || (nump < 1)) indata->eclbuf.ecloutbn = 5;
   }

  if (indata->eclbuf.ecloutbn == 3)
   {
    // get start and stop dates for central phase
    jd = spt[nump-1];
    for (i=0; i<nump; i++)
    if ((spt[i] < jd) && (spp[i] > 3)) jd = spt[i]; // start of central phase
    caldat(jd, dd, mm, yy, hh);
    dms (hh,deg,mnt,sec);
    sec = 0;
    i = mnt / indata->eclbuf.cstep;
    mnt = i*indata->eclbuf.cstep;   // cut to proper time step
    hh = ddd (deg, mnt, sec);
    jd = mjd (dd, mm, yy, hh);
    jd = jd - indata->tzone/24.0;      // use UT
    jdf = ept[nump-1];
    for (i=0; i<nump; i++)
     if ((ept[i] > jdf) && (spp[i] > 3)) jdf = ept[i]; // end of central phase
    jdf = jdf - indata->tzone/24.0;      // use UT

    strcpy(dcmp,"******");  // compare the date to know when to print it

    // get first output line for central eclipse
    EclCentral(1, p, jd, jdf, indata, dcmp, otxt);
    jd += double(indata->eclbuf.cstep) / (24.0*60.0);
   }

  if (indata->eclbuf.ecloutbn == 4)   // consecutive central eclipse data
   {
    EclCentral(0, p, jd, jdf, indata, dcmp, otxt);
    jd += double(indata->eclbuf.cstep) / (24.0*60.0);
    if (jd > jdf) indata->eclbuf.ecloutbn = 5;
    else indata->eclbuf.ecloutbn--;
   }

  if (indata->eclbuf.ecloutbn == 6)   // get maximum eclipse
   {
    jd = indata->eclbuf.eclmjd[j] - indata->tzone/24.0;
    EclCentral(0, p, jd, jdf, indata, dcmp, otxt);
    indata->eclbuf.ecloutbn = 0;
   }
 }

/***************************************************************************
 Get the date of Easter for the year yr as a text string otxt.
 ***************************************************************************/
void GetEaster (int yr, char* otxt)
{
 int c, n, k, i, j, l, d, m;
 double jd, h;
 char itxt[25];

 if (yr > 1582)  // Gregorian calendar
  {
   c =  yr / 100;
   n = yr / 19;
   n = yr - 19 * n;
   k = (c - 17) / 25;
   i = c / 4;
   j = (c - k) / 3;
   i = c - i - j + 19 * n +15;
   j = i / 30;
   i = i - 30 * j;
   j = i / 28;
   m = 29 /(i+1);
   n = (21 - n) / 11;
   i = i - j * (1 -j * m * n);
   j = yr / 4;
   m = c / 4;
   j = yr + j + i + 2 - c + m;
   m = j / 7;
   j = j - 7 * m;
   l = i - j;
   m = (l + 40) / 44 + 3;
   d = m / 4;
   d = l + 28 - 31 * d;
  }
 else             // Julian Calendar
  {
   c = yr % 4;
   n = yr % 7;
   k = yr % 19;
   i = (19*k +15) % 30;
   j = (2*c + 4*n - i +34) % 7;
   l = i + j + 114;
   m = l / 31;
   d = l % 31 + 1;
  }
 if (d > 31) d = 0;

 // calculate Ash Wednesday (46 days before Easter)
 jd = mjd(d, m, yr, 0.0) - 46.0;
 caldat(jd, c, n, k, h);
 sprintf(otxt,"Ash Wednesday:\t\t");
 GetMonth(n, itxt);
 strcat(otxt, itxt);
 sprintf(itxt,"%02i", c);
 strcat(otxt, itxt);

 strcpy(itxt, "\nEaster:\t\t\t");
 strcat(otxt, itxt);
 GetMonth (m, itxt);
 strcat(otxt, itxt);
 sprintf(itxt,"%02i", d);
 strcat(otxt, itxt);

}

/***************************************************************************
 Get the dates of Islamic New Year and First Day of Ramadan.
 ***************************************************************************/
double MJDIslam (int dd, int mm, int yr)
// Modified Julian Date of Islamic date given by day dd, month mm and year yr
{
 long int jd0;
 double result;

 // jd0 = 1948439; // value for astronomical use
 jd0 = 1948440;   // value for religious use
 jd0 = (11*yr+3)/30 + 354*yr + 30*mm - (mm - 1)/2 + dd + jd0 - 385;

 result = jd0 - 2400000.5;

 return result;
 }

void DJIslam (double md, int &d, int &m, int &y)
// Get Islamic Date (day d, month m and year y) from Modified Julian Date md
 {
  long int jd0, jd, l, n, j;

  jd = long(md + 2400001.0);
  jd0 = 1948440;

  l = jd - jd0 + 10632;
  n = (l - 1) / 10631;
  l = l - 10631*n + 354;
  j = ((10985 - l)/5316) * ((50*l)/17719) + (l/5670) * ((43*l)/15238);
  l = l - ((30 - j)/15) * ((17719*j)/50) - (j/16) * ((15238*j)/43) + 29;
  m = (24*l) / 709;
  d = l - (709*m) / 24;
  y = 30*n + j -30;
  if (m < 1) m = 1;
  if (m > 12) m = 12;
 }

void GetIslam (int yr, char* otxt)
// year yr
 {
  int d, m, y, yi, ya;
  double jy, h;
  char tlin[40];

  if (yr > 621)  // only defined from 622 on
   {
    h = 0;

    // get the Islamic New Year
    jy = mjd (1, 1, yr, h);
    DJIslam (jy, d, m, yi);

    if (m < 9) ya = yi;  // Ramadan is still in this year
    else ya = yi + 1;

    if ((m == 1) && (d == 1))
     {
      // it's Islamic New Year already! 
      y = yr;
     }
    else
     {
      yi+=1;
      d = 1;
      m = 1;
      jy = MJDIslam(d, m, yi);
      caldat (jy, d, m, y, h);
     };

    strcpy (otxt, "\n\nIslamic New Year:\t");
    GetMonth (m, tlin);
    strcat(otxt, tlin);
    sprintf(tlin,"%02i    (Islamic Year:%4i)", d, yi);
    strcat(otxt, tlin);

    // check for possible 2nd occurence
    if (m == 1)
     {
      yi+=1;
      d = 1;
      m = 1;
      jy = MJDIslam(d, m, yi);
      caldat (jy, d, m, y, h);
      if (y == yr)
       {
        strcat(otxt, "\n\t     and:\t");
        GetMonth (m, tlin);
        strcat(otxt, tlin);
        sprintf(tlin,"%02i    (Islamic Year:%4i)", d, yi);
        strcat(otxt, tlin);
       }
     }

    // now let's get Ramadan
    d = 1;
    m = 9;
    yi = ya;
    jy = MJDIslam(d, m, yi);
    caldat (jy, d, m, y, h);

    if (y > yr)
     {  // we went one Islamic year too far
      d = 1;
      m = 9;
      yi -= 1;
      jy = MJDIslam(d, m, yi);
      caldat (jy, d, m, y, h);
     };

    if (y < yr)
     {  // we are too early
      d = 1;
      m = 9;
      yi += 1;
      jy = MJDIslam(d, m, yi);
      caldat (jy, d, m, y, h);
     };

    strcat (otxt, "\nFirst Day Ramadan:\t");
    GetMonth (m, tlin);
    strcat (otxt, tlin);
    sprintf(tlin,"%02i", d);
    strcat(otxt, tlin);

    // check for possible 2nd occurence
    if (m == 1)
     {
      yi+=1;
      d = 1;
      m = 9;
      jy = MJDIslam(d, m, yi);
      caldat (jy, d, m, y, h);
      if (y == yr)
       {
        strcat(otxt, "\n\t     and:\t");
        GetMonth (m, tlin);
        strcat(otxt, tlin);
        sprintf(tlin,"%02i", d);
        strcat(otxt, tlin);
       }
     }
   }
  else strcpy (otxt, " ");
 }

/***************************************************************************
 Get the date of the Indian New Year.
 ***************************************************************************/
double MJDIndia (int dd, int mm, int yr)
// Modified Julian Date of Indian date given by day dd, month mm and year yr
{
 long int jd;
 double result;

 jd = 365*yr + (yr + 78 - 1/mm) / 4 + 31*mm - (mm + 9)/11 - (mm/7)*(mm - 7)
             - (3*((yr + 78 - 1/mm) / 100 + 1)) / 4 + dd + 1749579;

 result = jd - 2400000.5;

 return result;
}

void DJIndia (double md, int &d, int &m, int &y)
// Get Indian Date (day d, month m and year y) from Modified Julian Date md
 {
  long int jd, l, n, i, j;

  jd = long(md + 2400001.0);
  l = jd + 68518;
  n = (4*l) / 146097;
  l = l - (146097*n + 3) / 4;
  i = (4000*(l + 1)) / 1461001;
  l = l - (1461*i) / 4 + 1;
  j = ((l-1)/31)*(1-l/185)+(l/185)*((l-156)/30+5)-l/366;
  d = l - 31*j + ((j+2)/8)*(j-5);
  l = j/11;

  m = j + 2 - 12*l;
  y = 100*(n-49) + l + i - 78;
  if (m < 1) m = 1;
  if (m > 12) m = 12;
 }

void GetIndia (int yr, char* otxt)
// year yr
 {
  int d, m, y, yi;
  double jy, h;
  char tlin[40];

  // get the Indian New Year
  if (yr > 81)    // algorithm not defined before
   {
    h = 0;
    jy = mjd (1, 1, yr, h);
    DJIndia (jy, d, m, yi);
    d = 1;
    m = 1;
    yi+=1;
    jy = MJDIndia (d, m, yi);
    caldat (jy, d, m, y, h);

    strcpy (otxt, "\n\nIndian New Year:\t");
    GetMonth (m, tlin);
    strcat(otxt, tlin);
    sprintf(tlin,"%02i    (Indian Year:%4i)", d, yi);
    strcat(otxt, tlin);
   }
  else strcpy (otxt, " ");
 }

/***************************************************************************
 Get the date of the Chinese New Year.
 ***************************************************************************/
void ChinYName (int year, char *tlin)
 {
  /* get the name of the Chinese Year
     year : Gregorian year when the respective Chinese year starts
     tlin : text output for the name of the year
  */
  int yr;
  char ttx[15];
  char celst[10][5] = {"jia","yi","bing","ding","wu","ji","geng",
                                   "xin","ren","gui"};
  char earthb[12][13] = {"zi (rat","chou (ox","yin (tiger","mao (hare",
                              "chen (dragon","si (snake","wu (horse",
                              "wei (sheep","shen (monkey","you (fowl",
                              "xu (dog","hai (pig"};

  yr = (year-1984) % 10;
  if (yr < 0) yr += 10;
  strcpy(tlin,celst[yr]);

  yr = (year-1984) % 12;
  if (yr < 0) yr += 12;
  strcpy(ttx,earthb[yr]);
  strcat(tlin,"-");
  strcat(tlin,ttx);
  strcat(tlin,")");
 }

double sunlong (double jd, double tzone, double deltdut)
 {
  /* Calculate the ecliptic longitude of the Sun
     jd : MJD of the time for the calculation
     tzone : time zone (in hours) of jd
     deltdut : difference TDT - UT in sec

     RETURN: ecliptic longitude of the Sun in degrees
             normalized between 180 and 540 degrees
  */

  Sun200 sun;
  Vec3 rs;
  double jc;

  jc = jd - tzone/24.0;
  jc = jc + deltdut/86400.0;
  jc = julcent(jc);
  rs = sun.position(jc);
  rs = carpol(rs);
  jc = rs[1] * 180.0 / M_PI;
  if (jc < 0) jc += 360.0;
  if (jc < 180.0) jc += 360.0;

  return jc;
 }

double JDChNY (WinpData* indata, WoutData* outdata, long int *newm, int& mp0)
 {
  /* Get the MJD of the Chinese New Year.
     newm contains the New Moon MJD's from one winter solstice to
     the next.
     mp0 is the number of months in that year.
  */
  double savtz, yr1, yr2, ejd, emag, td;
  double pd, pl;
  int eph, icflg;
  long int winter1, winter2, mnm;

  // save current values
  savtz = indata->tzone;

  indata->tzone = 8.0;  // longitude 120 deg (Beijing)
  indata->year--;
  outdata->curlines = 0;
  moonph(indata, outdata);
  winter1 = long(outdata->wsolstice);  // winter solstice the year before
  indata->year++;
  outdata->curlines = 0;
  moonph(indata, outdata);
  winter2 = long(outdata->wsolstice); // next winter solstice

  // calculate the New Moons between the winter solstices
  mp0 = 0;
  yr1 = indata->year - 0.25;
  yr2 = indata->year + 1.25;

  while (yr1 < yr2)
   {     /* calculate New Moons*/
    td = phmjd (yr1, 0.0, indata->del_tdut, eph, ejd, emag);
    td = td + indata->tzone/24.0;
    td = td - indata->del_tdut/86400.0;
    mnm = long(td);
    if (mnm <= winter1) newm[0] = mnm;
    else
     {
      if ((mnm <= winter2) && (mp0 < 15))
       {
        if (mnm > newm[mp0])
         {
          mp0++;
          newm[mp0] = mnm;
         };
       };
     };
    yr1+=0.02;  /* incease by 1/50th of a year (about two weeks) */
   };

  icflg = 0;  // no intercalary month
  pl = 300.0;
  if (mp0 > 12)  // check for possible intercalary month
   {
    pd = double(newm[1]);
    if (sunlong (pd, indata->tzone, indata->del_tdut) > pl) pl = 330.0;
    pd = double(newm[2]);
    if (sunlong (pd, indata->tzone, indata->del_tdut) < pl) icflg = 1;
    else pl += 30.0;
    pd = double(newm[3]);
    if (sunlong (pd, indata->tzone, indata->del_tdut) < pl) icflg = 1;
   }

  if (icflg == 0) td = double (newm[2]);  // normal year
  else td = double (newm[3]);  // intercalary month in between

  // restore timezone
  indata->tzone = savtz;

  return td;
 }

void GetChina (WinpData* indata, WoutData* outdata, char *otxt)
 {
  // Get the date of the Chinese New Year and the name of the year.
  // Output the result in otxt.
	
  char tlin[40];
  double yr2, td;
  int mp0, eph, y;
  long int newm[16];

  td = JDChNY (indata, outdata, newm, mp0);

  strcpy (otxt, "\n\nChinese New Year:\t");

  caldat (td, eph, mp0, y, yr2);
  GetMonth (mp0, tlin);
  strcat(otxt, tlin);
  sprintf(tlin,"%02i\t", eph);
  strcat(otxt, tlin);

  ChinYName(indata->year, tlin);
  strcat (otxt, tlin);

  // restore values
  outdata->curlines = 0;
  moonph(indata, outdata);
 }

/***************************************************************************
 Get the date of Jewish New Year for the year yr as a text string otxt.
 ***************************************************************************/

/*------------------------- GetTishMol ---------------------------------*/

double GetTishMol (WinpData* indata, int& jyr)
{
  /* get the time since Tishri molad of A.M. 1 for the
     year indata-year
     Also return the Jewish year jyr.
  */
 int c19, noy, nly, oyflg;
 long int halakim, hours, days, weeks, ovf;
 double ckh, jd;

 jyr = indata->year + 3760;  // Jewish year - 1
 c19 = jyr / 19;  // # of 19 year cycles
 noy = jyr % 19;
 nly = 0;        // # of leap years
 if (noy > 2) nly = 1;
 if (noy > 5) nly = 2;
 if (noy > 7) nly = 3;
 if (noy > 10) nly = 4;
 if (noy > 13) nly = 5;
 if (noy > 16) nly = 6;
 if (nly > 0) noy -= nly;  // # of ordinary years

 halakim = c19*595 + noy*876 + nly*589 + 204;
 ovf = halakim / 1080;
 halakim %= 1080;

 hours = c19*16 + noy*8 + nly*21 + 5 + ovf;
 ovf = hours / 24;
 hours %= 24;

 days = c19*2 + noy*4 + nly*5 + 2 + ovf;
 ovf = days % 7;   // weekday of Tishri molad

 weeks = c19*991 + noy*50 + nly*54;
 days = days + weeks*7;   // total number of days since Tishri molad 1

 // now figure out the dehiyyot
 oyflg = jyr % 19;
 if ((oyflg == 2) || (oyflg == 5) || (oyflg == 7) || (oyflg == 10)
      || (oyflg == 13) || (oyflg == 16) || (oyflg == 18)) oyflg = 1;
 else oyflg = 0;  // this is an ordinary year

 nly = 0;  // additional days
 if ((oyflg == 0) && (ovf == 3))  // check rule (c)
  {
   ckh = double (halakim);
   ckh = double (hours) + ckh / 1080.0;
   if (ckh > 9.18888888889) nly = 2;
  };

 if (nly == 0)
  {
   oyflg = (jyr-1) % 19;
   if ((oyflg == 2) || (oyflg == 5) || (oyflg == 7) || (oyflg == 10)
           || (oyflg == 13) || (oyflg == 16) || (oyflg == 18)) oyflg = 1;
   else oyflg = 0;  // was preceding year leap?
   if ((oyflg) && (ovf == 2))   // check rule (d)
    {
     ckh = double (halakim);
     ckh = double (hours) + ckh / 1080.0;
     if (ckh > 15.5453703704) nly = 1;
    };
   };

  if (nly == 0)
   {
    if (hours >= 18)  // check rule (b)
     {
      nly = 1;
      ovf++;
      ovf %= 7;
     };
    if ((ovf == 1) || (ovf == 4) || (ovf == 6)) nly++;  // rule (a)
   };

  days = days + nly - 2052005;
  jd = double (days);  // MJD of Tishri molad

  return jd;
 }

/*------------------------- GetJewish ---------------------------------*/

void GetJewish (WinpData* indata, char* otxt)
{
 int jyr, c19, noy, nly;
 double ckh, jd;
 char itxt[31];

 if (indata->year < -3762)  // only calculate for Jewish Years >= 0
  {
   strcpy (otxt, " ");
   return; 
  };
  
 // get the time since Tishri molad of A.M. 1
  jd = GetTishMol (indata, jyr);

  strcpy (otxt, "\n\nJewish New Year:\t");
  caldat (jd, noy, nly, c19, ckh);
  GetMonth (nly, itxt);
  strcat(otxt, itxt);
  jyr++;
  sprintf(itxt,"%02i    (Jewish Year:%4i)", noy, jyr);
  strcat(otxt, itxt);

  // now get Yom Kippur
  jd += 9.0;
  strcpy (itxt, "\nYom Kippur:\t\t");
  strcat (otxt, itxt);
  caldat (jd, noy, nly, c19, ckh);
  GetMonth (nly, itxt);
  strcat(otxt, itxt);
  sprintf(itxt,"%02i", noy);
  strcat(otxt, itxt);
 }

/*********************************************************************/

/*------------------------- HebrewDM ---------------------------------*/

void HebrewDM (double jd, double jd1, int lyr, int &d, int &m, int &j)
 {
  /* Get the day and month of the Hebrew calendar for
     the Modified Julian Date jd.
     jd1 = MJD of last Jewish New Year
     lyr = lenght of Jewish year
     d = day number
     m = month number
     j = month index
  */
  int dflg;
  int k, n;
  
  d = int(jd - jd1 + 1);
  dflg = 1;
  j = 0;
  m = 1;

  for (k=1; k<13; k++)
   {
    if (dflg)
     {
      // length of month
      if (k % 2) n = 30;
      else n = 29;
      if ((k==2) && ((lyr==355) || (lyr==385))) n = 30;  // complete year
      if ((k==3) && ((lyr==353) || (lyr==383))) n = 29;  // deficient year
      if ((k==6) && (lyr > 360))   // leap year
       {
        if (d <= 30)  // Adar I
         {
          m = k;
          dflg = 0;
          j = 12;
         }
        else d -= 30;
       };

      if (dflg)
       {
        if (d <= n)
         {
          m = k;
          dflg = 0;
         }
        else
         {
          d -= n;
          j++;
         };
       };
     };
   };

  if (j > 13) j = 0;  // just in case
  if ((m==6) && (lyr > 360))   // check for Adar II
   {
    if (j != 12) j = 13;
   };
 }

/*------------------------- ChineseDM ---------------------------------*/

void ChineseDM (WinpData* indata, double jd, double jd1, long int *newm,
                                  int lyr, int &d, int &m, int &j)
 {
  /* Get the day, month and year for the Chinese calendar for
     the Modified Julian Date jd.

     jd1 = MJD of Chinese New Year
     newm = New Moon dates (MJD) from winter solstice to winter solstice
     lyr = number of months in current Chinese year
     d = day number
     m = month number  (negative if intercalary)
     j = corresponding Gregorian year number needed for the
         Chinese year name
   */
  int cny, icm, k, icflg;
  double pl, pd;

  // get the index of the New Moon marking the Chinese New Year
  if (jd1 == double(newm[2])) cny = 2;
  else cny = 3;

  // find the index of the intercalary month in a leap year
  icm = -1;
  if (lyr > 12)
   {
    pl = 300.0;   // nominal longitude to define principal term
    pd = double(newm[1]);
    if (sunlong (pd, indata->tzone, indata->del_tdut) > pl) pl += 30;
    for (k = 2; k<= lyr; k++)
     {
      if (icm < 0)   // icm is the index for the intercalary month
       {
        pd = double(newm[k]);
        if (sunlong (pd, indata->tzone, indata->del_tdut) <= pl) icm = k - 1;
        else pl += 30.0;
        if (pl > 540.0) pl -= 360.0;
        if (sunlong (pd, indata->tzone, indata->del_tdut) > pl) pl += 30;
        if (pl > 540.0) pl -= 360.0;
       };
     };
    if (icm > lyr) icm = lyr;  // just in case
   };

  // find the index of the month of the respective day
  m = -1;
  for (k = 1; k <= lyr; k++)
   {
    pd = double(newm[k]);
    if (m < 0) if (jd < pd) m = k - 1;
   };
  if (m < 0) m = 0;
  if (m > lyr) m = 0;  // this shouldn't happen - just in case!

  // now assign the day, month, year
  d = int(jd - double(newm[m]) + 1);
  j = indata->year;
  if (m < cny) j--;

  icflg = 0;
  if (icm > 0)  // special case of leap year
   {
    if (m == icm) icflg = 1;
    if (m >= icm) m--;
   };
  if (m < 2) m += 11;
  else m --;
  if (icflg) m = -m;
 }

/*------------------------- CompDayDet ---------------------------------*/

void CompDayDet (WinpData* indata, WoutData* outdata, char* otxt, int k)
 {
  /* Compute the details of the respective day e.g. which date
     it is in other calendars, which day of the year and which
     week of the year.
     k is 1 for the first call, 2 for the second call
  */
  char mtxt[50];
  char wkday[7][10] = {"Monday","Tuesday","Wednesday",
                       "Thursday","Friday","Saturday","Sunday"};
  char mislam[12][14] = {"Muharram", "Safar", "Rabi'a I",
                         "Rabi'a II", "Jumada I", "Jumada II",
                         "Rajab", "Sha'ban", "Ramadan",
                         "Shawwal", "Dhu al-Q'adah", "Dhu al-Hijjah"};
  char mindian[12][11] = {"Caitra", "Vaisakha", "Jyaistha", "Asadha",
                          "Stavana", "Bhadra", "Asvina", "Kartika",
                          "Agrahayana", "Pausa", "Magha", "Phalguna"};
  char mhebrew[14][8] = {"Tishri", "Heshvan", "Kislev", "Tevet",
                         "Shevat", "Adar", "Nisan", "Iyar", "Sivan",
                         "Tammuz", "Av", "Elul", "Adar I", "Adar II"};

  double jd, jd0, jd1, jd2;
  int j, d, m, y, jyr, lyr;
  long int newm[16];

  jd = mjd(indata->day, indata->month, indata->year,0.0);

  if (k == 1)
   {
    // get weekday
    j = int(jd);
    j += 2;
    if (j < 0)
     {
      j = -1 - j;
      j %= 7;
      j = 6 - j;
     }
    else j %= 7;  // first weekday of the month (0 = Monday, ... 6 = Sunday)
    strcpy (otxt, wkday[j]);

    // get the rest of the (Gregorian/Julian) day
    sprintf(mtxt, "   %2i-", indata->day);
    strcat(otxt, mtxt);
    GetMonth (indata->month, mtxt);
    strcat (otxt, mtxt);
    sprintf(mtxt, "-%4i", indata->year);
    strcat(otxt, mtxt);

    // day of year and week#
    jd0 = mjd(1, 1, indata->year, 0.0);
    d = int(jd - jd0 + 1);
    sprintf(mtxt, "\n\nDay of Year: %3i", d);
    strcat(otxt, mtxt);

    // get weekday of first day of year
    j = int(jd0);
    j += 2;
    if (j < 0)
     {
      j = -1 - j;
      j %= 7;
      j = 6 - j;
     }
    else j %= 7;  // first weekday of the month (0 = Monday, ... 6 = Sunday)
    if (j > 3) y = d - 8 + j;
    else y = d + j - 1;
    if (y >= 0) y = y/7 + 1;
    else y = 0;

    if (y > 0)
     {
      sprintf(mtxt,"     Week: %2i", y);
      strcat(otxt, mtxt);
     };
    strcat(otxt,"\n======================================\n");

    // Jewish day
    // get the time since Tishri molad of A.M. 1 for current year
    if (indata->year > -3763)
     {
      jd0 = GetTishMol (indata, jyr);
      y = jyr;
      // get the second Tishri molad bracketing the date
      if (jd0 == jd)  // we've already found the date
       {
        d = 1;
        m = 1;
        lyr = 354;
        j = 0;
       }
      else
       {
        d = indata->year;
        if (jd0 > jd) indata->year--;
        else indata->year++;
        jd1 = GetTishMol (indata, jyr);
        indata->year = d;
        jd2 = jd0;
        if(jd2 < jd1)
         {
          jd0 = jd1;
          jd1 = jd2;
          jd2 = jd0;
         }  // now jd1 is the first and jd2 the next Tishri molad
        else y = jyr;
        y++;    // y is now the correct Jewish year.
        jd0 = jd2 - jd1;
        lyr = int(jd0);    // length of year
        HebrewDM (jd, jd1, lyr, d, m, j);
       };
      sprintf(mtxt, "\n\nHebrew Day: %2i     Month %2i",d,m);
      strcat(otxt, mtxt);
      if ((lyr > 360) && (j == 13)) strcat(otxt, "+ ("); // leap year intercalary
      else strcat(otxt, " (");
      strcat(otxt, mhebrew[j]);
      sprintf(mtxt, ")    Year: %4i", y);
      strcat(otxt, mtxt);
     };  

    // Islamic day
    if (indata->year > 621)
     {
      DJIslam (jd, d, m, y);
      sprintf(mtxt, "\n\nIslamic Day: %2i    Month %2i (",d, m);
      strcat(otxt, mtxt);
      strcat(otxt, mislam[m-1]);
      sprintf(mtxt, ")    Year: %4i", y);
      strcat(otxt, mtxt);
     };
   };

  if (k == 2)
   {
    // Indian day
    if (indata->year > 81)
     {
      DJIndia (jd, d, m, y);
      sprintf(otxt, "\n\nIndian Day: %2i    Month %2i (",d, m);
      strcat(otxt, mindian[m-1]);
      sprintf(mtxt, ")    Year: %4i", y);
      strcat(otxt, mtxt);
     }
    else strcpy(otxt," ");

    // Chinese day
    jd0 = indata->year;
    jd1 = JDChNY (indata, outdata, newm, lyr);
    jd2 = double(newm[lyr]);
    if (jd >= jd2)
     {
      indata->year++;
      jd1 = JDChNY (indata, outdata, newm, lyr);
     };
    ChineseDM (indata, jd, jd1, newm, lyr, d, m, j);

    if (indata->year > -4712)
     {
      sprintf(mtxt, "\n\nChinese Day: %2i    Month %2i ",d, abs(m));
      strcat (otxt, mtxt);
      if (m < 0) strcat (otxt, "(intercalary) ");
      strcat (otxt, "   Year: ");
      ChinYName(j, mtxt);
      strcat (otxt, mtxt);
     }; 

    // restore values
    indata->year = int(jd0);
    outdata->curlines = 0;
    moonph(indata, outdata);

    strcat(otxt,
        "\n\n\nJewish and Islamic days begin at sunset on the previous evening");
   };
 }

/*********************************************************************/

/*--------------------------- SnPos ---------------------------------*/

Vec3 SnPos (WinpData* indata, double jd, double &ep2, double &els)
 {
  // return the apparent position of the Sun
  // and the Nutation ep2 value and the ecliptic longitude of the Sun els

  Sun200 sun;
  Mat3 mx;
  Vec3 rs, s;
  double t;

  if (indata->del_auto) t = DefTdUt(indata->year);
  else t = indata->del_tdut;
  t = jd + t/86400.0;
  t = julcent (t);

  rs = sun.position(t);
  s = carpol(rs);
  els = s[1];

  rs = eclequ(t,rs);
  mx = nutmat(t,ep2);   // nutation
  rs = mxvct(mx,rs);    // apparent coordinates
  rs = aberrat(t,rs);

  return rs;
 }

/*--------------------------- MnPos ---------------------------------*/

Vec3 MnPos (WinpData* indata, double jd, double &ep2, double &elm)
 {
  // return the apparent position of the Moon
  // and the Nutation ep2 value and the ecliptic longitude of the Moon elm
	
  Moon200 moon;
  Mat3 mx;
  Vec3 rm, s;
  double t;

  if (indata->del_auto) t = DefTdUt(indata->year);
  else t = indata->del_tdut;
  t = jd + t/86400.0;
  t = julcent (t);

  rm = moon.position(t);
  s = carpol(rm);
  elm = s[1];

  rm = eclequ(t,rm);
  mx = nutmat(t,ep2);   // nutation
  rm = mxvct(mx,rm);

  return rm;
 }

/*------------------------- RiseSet ---------------------------------*/

int RiseSet (double jd, Vec3 rs, double ht, double elv, double ep2,
                                 double lat, double lng, double &jdr, double &jds)
 {
  /* Get the MJD of "rise" and "set" times for the object with
     equatorial position vector rs at MJD time jd.
     The event is characterized by an elevation angle of ht (in degrees).
     elv = elevation of body at time jd (in degrees)
     ep2 = correction factor for nutation
     lat = geographic latitude (in radians)
     lng = geographic longitude (in degrees)
     jdr = MJD of the "rise" event
     jds = MJD of the "set" event

     RETURN: 1 if rise and set times were found
             0 if object always above the horizon
             -1 if object never above the horizon
  */
  double h0, dec, ra, dn, ts;
  const double tpi = 2.0 * M_PI;
  int mflg, elflg;
  Vec3 s;

  if (elv >= ht) elflg = 0;
  else elflg = -1;

  h0 = sin(ht*M_PI/180.0);
  s = carpol(rs);  // get right ascension and declination
  dec = s[2];
  ra =  s[1];
  mflg = 1;
  h0 = h0 - sin(lat) * sin(dec);
  dn = cos(lat) * cos(dec);
  if (dn == 0) mflg = 0;
  else
   {
    h0 /= dn;
    if (fabs(h0) > 1.0) mflg = 0;
   };

  if (mflg)
   {
    ts = ra / tpi - lsidtim (jd, lng, ep2) / 24.0;
    if (ts < 0) ts = ts + 1.0;
    if (ts > 1.0) ts = ts - 1.0;
    h0 = acos(h0) / tpi;
    jdr = jd + ts - h0;
    jds = jd + ts + h0;
   }
  else mflg = elflg;

  return mflg;
 }

/*------------------------- TwiOut ---------------------------------*/

void TwiOut (double jdr, double jds, double tz, int d, int g,
				 char *otxt, char *mtxt)
 {
  /* prepare the output for the beginning and end of twilight times
     jdr = MJD of beginning of twilight
     jds = MJD of end of twilight
     d = 1 if rise and set times were found
         0 if object always above the horizon
         -1 if object never above the horizon
     g = angle of twilight
     otxt, mtxt text output buffers
  */

  if (d == 1)
   {
    jdr += tz/24.0;
    dtmstr (jdr, mtxt);
    mtxt[12]= '\0';
    strcat (otxt, mtxt);
    strcat (otxt, "  -  ");
    jds += tz/24.0;
    dtmstr (jds, mtxt);
    mtxt[12]= '\0';
   }
  else
   {
    if (d < 0) sprintf (mtxt, "Sun always below %i deg", g);
    else sprintf (mtxt, "Sun always above %i deg", g);
   };
  strcat (otxt, mtxt);
 }

/*------------------------- linpol ---------------------------------*/

double linpol(double t1, double t2, double e1, double e2, double ex)
 {
  /* linear interpolation between points (t1, e1) and (t2, e2).

     RETURN: t for which the ordinate is ex.
  */
  double al;

  al = (e2 - e1) / (t2 - t1);
  if (fabs(al) > 1e-180) al = 1.0 / al;
  else al = 0;
  al = (ex - e1) * al + t1;

  return al;
 }

/*------------------------- Twiacc ---------------------------------*/

double Twiacc (WinpData* indata, double jdr, double ep2, double lat,
                                 double lng, double twa)
 {
  // Increase the accuracy of the sun rise, set and twilight data
  // by linear interpolation
	
  double ae = 23454.77992; // 149597870.0/6378.14 =  1AE -> Earth Radii
  Vec3 rs;
  double jd1, jd2;
  double azim, el1, el2, dist, ps;

  jd1 = jdr;
  jd2 = jdr + 0.2 / 24.0;
  rs = SnPos (indata, jd1, ep2, ps);
  rs *= ae;
  AppPos (jd1, ep2, lat, lng, indata->geoheight, 1, rs,
                    azim, el1, dist);
  rs = SnPos (indata, jd2, ep2, ps);
  rs *= ae;
  AppPos (jd2, ep2, lat, lng, indata->geoheight, 1, rs,
                    azim, el2, dist);
  jd1 = linpol(jd1, jd2, el1, el2, (twa*degrad));

  // limit the adjustment just in case
  if (fabs(jdr - jd1) > 0.125)   // 18 min difference
   {
    if (jd1 > jdr) jd1 = jdr + 0.125;
    else jd1 = jdr - 0.125;
   };

  return jd1;
 };

/*------------------------- MoonLibr ---------------------------------*/

void MoonLibr (double jd, Vec3 rm, Vec3 sn,
               double &lblon, double &lblat, double &termt)
 {
  /* Calculate the librations angles lblon (longitude) and
     lblat (latitude) for the moon at position rs and time jd.
     Also calculate the selenographic longitude of the terminator.
     rm is the position of the Moon from Earth, sn the position
     of the Sun (also with respect to Earth).
  */
  double t, gam, gmp, l, omg, mln;
  double a, b, c, ic, gn, gp, omp;
  double const degrad = M_PI / 180.0;
  Vec3 v1;
  Mat3 m1, m2;

  t = (jd - 15019.5) / 36525.0;
  gam = 281.2208333 + ((0.33333333e-5*t + 0.45277778e-3)*t + 1.7191750)*t;
  gmp = 334.3295556 + ((-0.125e-4*t - 0.10325e-1)*t + 4069.0340333)*t;
  l = 279.6966778 + (0.3025e-3*t + 36000.768925)*t;
  omg = 259.1832750 + ((0.22222222e-5*t + 0.20777778e-2)*t - 1934.1420083)*t;
  b = 23.45229444 + ((0.50277778e-6*t -0.16388889e-5)*t - 0.130125e-1)*t;
  mln = 270.4341639 + ((0.1888889e-5*t -0.11333e-2)*t + 481267.8831417)*t;
  ic = 1.535*degrad;
  gn = (l - gam)*degrad;
  gp = (mln - gmp)*degrad;
  omp = (gmp - omg)*degrad;
  a = -107.0*cos(gp) + 37.0*cos(gp+2.0*omp) -11.0*cos(2.0*(gp+omp));
  a = a*0.000277778*degrad + ic;
  c = (-109.0*sin(gp) + 37.0*sin(gp+2.0*omp) - 11.0*sin(2.0*(gp+omp)))/sin(ic);
  c = (c*0.000277778 + omg)*degrad;
  gn = -12.0*sin(gp) + 59.0*sin(gn) + 18.0*sin(2.0*omp);
  gn = gn*0.000277778*degrad;  // tau

  b *= degrad;
  gam = cos(a)*cos(b) + sin(a)*sin(b)*cos(c);
  gmp = gam*gam;
  if(gmp > 1.0) gmp = 0;
  else gmp = sqrt(1.0 - gmp);
  gam = atan23(gmp,gam);  // theta
  t = cos(a)*sin(b) - sin(a)*sin(b)*cos(c);
  l = -sin(a)*sin(c);

  gmp = atan23(l,t);  // phi
  t = sin(a)*cos(b) - cos(a)*sin(b)*cos(c);
  l = -sin(b)*sin(c);
  a = atan23(l,t);  // delta
  c = a + mln*degrad + gn - c;   // psi

  // libration rotation matrix from Mean equator to true selenographic
  m1 = zrot(gmp);
  m2 = xrot(gam);
  m1 = m2 * m1;
  m2 = zrot(c);
  m1 = m2 * m1;

  // Earth coordinates
  v1[0] = -rm[0];
  v1[1] = -rm[1];
  v1[2] = -rm[2];

  v1 = mxvct(m1,v1);
  v1 = carpol(v1);
  lblat = v1[2] / degrad;
  lblon = v1[1] / degrad;
  if (lblon > 180.0) lblon = lblon - 360.0;

  // terminator
  v1 = mxvct(m1,sn);
  v1 = carpol(v1);
  termt = v1[1] / degrad;
  if (termt > 180.0) termt = termt - 360.0;
  termt -= 90.0;
  a = 90.0 + lblon;
  b = lblon - 90;
  if (termt > a) termt -= 180.0;
  else
   {
    if (termt < b) termt += 180;
   };
 }

/*------------------------- CompSkyDet ---------------------------------*/

void CompSkyDet (WinpData* indata, char* otxt, int k)
 {
  // Compute the Position of the Sun and the Moon for a specified
  // day and time.

  char mtxt[50];
  int d, m, rfnd;
  double jd, t, hour, lat, lng, lblon, lblat, termt, mnmag;
  double jdr, jds, jd1, jd2;
  double azim, elev, elsv, dist, hpar, elr, ps;
  static double els, elm;
  double const degrad = M_PI / 180.0;
  double ae = 23454.77992; // 149597870.0/6378.14 =  1AE -> Earth Radii
  static Vec3 sg, rg, snc, mnc;   // position of the Sun and the Moon
  Vec3 s, rm, rs, s3;
  double ep2;    // correction for Apparent Sideral Time

  hour = ddd(indata->hour, indata->minute, double(indata->second))
                           - indata->tzone;
  jd = mjd(indata->day, indata->month, indata->year,hour);

  if (k == 1)    // header output, Julian Date and Sidereal Time
   {
    strcpy(otxt, "Sky Circumstances for ");
    strncat(otxt, indata->geoname, 30);
    if (indata->dmsflg)  // deg.fff -> DD.MMSSS
     {
      lat = DegFDms(indata->geolat);
      lng = DegFDms(indata->geolong);
     }
    else
     {
      lat = indata->geolat;
      lng = indata->geolong;
     };
    sprintf(mtxt,"\nLat: %g   Long: %g   height: %g m\n\n",
                              lat, lng, indata->geoheight);
    decdot0(mtxt);
    strcat(otxt,mtxt);
    sprintf(mtxt,"%2i-", indata->day);
    strcat(otxt,mtxt);
    GetMonth (indata->month, mtxt);
    strcat (otxt, mtxt);
    sprintf(mtxt, "-%4i   ", indata->year);
    strcat(otxt, mtxt);
    sprintf(mtxt,"%2i:%2i:%2i",indata->hour, indata->minute, indata->second);
    strcat(otxt, mtxt);
    sprintf(mtxt,"    Timezone: %g", indata->tzone);
    decdot(mtxt);
    strcat(otxt, mtxt);
    t = jd + 2400000.5;
    sprintf(mtxt,"\n\nJulian Date:%14.4lf", t);
    decdot0(mtxt);
    strcat(otxt, mtxt);
    t = lsidtim (jd, indata->geolong, 0);
    if (t < 0) t += 24.0;
    t = DegFDms(t);
    sprintf(mtxt,"\nLocal Mean Sidereal Time (H.MMSS):%9.4lf\n\n",t);
    decdot0(mtxt);
    strcat(otxt, mtxt);
   };

  lat = indata->geolat * degrad;
  lng = indata->geolong * degrad;

  if ((k == 2) || (k == 3))  // Sun position and twilight
   {
    rs = SnPos (indata, jd, ep2, els);
    rs *= ae;
    snc = rs;
    sg = rs - GeoPos(jd, ep2, lat, lng, indata->geoheight);  // topocentric
    s = carpol(sg);
    dist = s[0];
    elev = s[2] / degrad;
    azim = s[1] / degrad;
    azim /= 15.0;
    if (azim < 0) azim += 24.0;
    dms (azim, d, m, hour);
    sprintf (otxt, "SUN:\n\nR.A. %2ih %2im %2.0fs   ", d, m, hour);
    decdot0(otxt);
    dms (elev, d, m, hour);
    sprintf (mtxt, "Dec. %2id %2im %2.0fs   ", d, m, hour);
    decdot0(mtxt);
    strcat(otxt, mtxt);

    AppPos (jd, ep2, lat, lng, indata->geoheight, 1, rs, azim, elev, dist);
    azim /= degrad;
    if (elev > -3.5e-2) elr = (elev + Refract(elev)) / degrad;
    else elr = elev / degrad;
    elev /= degrad;

    if (azim < 0) azim += 360.0;
    sprintf (mtxt, "    Azimuth%6.1lf   Elev.%6.1lf ", azim, elr);
    decdot0(mtxt);
    strcat(otxt, mtxt);
   };
  
  if (k == 3)  //  rise, set and twilight
   {
    // get rise and set
    strcpy (otxt, "\n");
    d = RiseSet (jd, rs, -0.83, elev, ep2, lat, indata->geolong, jdr, jds);
    if ((d == 1) && (jdr > jd))
     {
      jd -= 1.0;
      rs = SnPos (indata, jd, ep2, ps);
      rs *= ae;
      d = RiseSet (jd, rs, -0.83,
                   elev, ep2, lat, indata->geolong, jdr, jds);
     };

    if (d == 1)   // get dates more accurately
     {
      jdr = Twiacc (indata, jdr, ep2, lat, lng, -0.83);
      jds = Twiacc (indata, jds, ep2, lat, lng, -0.83);
     };

    if (d == 1)
     {
      strcat (otxt, "Rise: ");
      jdr += indata->tzone/24.0;
      dtmstr (jdr, mtxt);
      mtxt[12]= '\0';
      strcat (otxt, mtxt);
      strcat (otxt, "     Set: ");
      jds += indata->tzone/24.0;
      dtmstr (jds, mtxt);
      mtxt[12]= '\0';
      strcat (otxt, mtxt);
     }
    else
     {
      if (d < 0) strcat (otxt, "Sun always below the horizon");
      else strcat (otxt, "Sun always above the horizon");
     };

    // get twilight
    if (d != 0)
     {
      d = RiseSet (jd, rs, -6.0, elev, ep2, lat, indata->geolong, jdr, jds);
      if (d == 1)   // get dates more accurately
       {
        jdr = Twiacc (indata, jdr, ep2, lat, lng, -6.0);
        jds = Twiacc (indata, jds, ep2, lat, lng, -6.0);
       };
      strcat (otxt, "\nCivil Twilight            : ");
      TwiOut (jdr, jds, indata->tzone, d, -6, otxt, mtxt);
     };

    // get nautical twilight
    if (d != 0)
     {
      d = RiseSet (jd, rs, -12.0, elev, ep2, lat, indata->geolong, jdr, jds);
      if (d == 1)   // get dates more accurately
       {
        jdr = Twiacc (indata, jdr, ep2, lat, lng, -12.0);
        jds = Twiacc (indata, jds, ep2, lat, lng, -12.0);
       };
      strcat (otxt, "\nNautical Twilight      : ");
      TwiOut (jdr, jds, indata->tzone, d, -12, otxt, mtxt);
     };

    // get astronomical twilight
    if (d != 0)
     {
      d = RiseSet (jd, rs, -18.0, elev, ep2, lat, indata->geolong, jdr, jds);
      if (d == 1)   // get dates more accurately
       {
        jdr = Twiacc (indata, jdr, ep2, lat, lng, -18.0);
        jds = Twiacc (indata, jds, ep2, lat, lng, -18.0);
       };
      strcat (otxt, "\nAstronomical Twilight: ");
      TwiOut (jdr, jds, indata->tzone, d, -18, otxt, mtxt);
     };
   };

  if (k == 4)  // Moon position and sun angle
   {
    rm = MnPos (indata, jd, ep2, elm);
    mnc = rm;
    rg = rm - GeoPos(jd, ep2, lat, lng, indata->geoheight);  // topocentric
    s = snc - rm;
    MoonLibr(jd, rg, s, lblon, lblat, termt);  // librations and terminator

    // calculate apparent magnitude of the Moon
    mnmag = abs(rg) / 23454.77992;  // distance moon-observer in AU
    s = vnorm(s);
    s[0] = -s[0];
    s[1] = -s[1];
    s[2] = -s[2];
    s3 = vnorm(rg);

    dist = dot(s, s3);  // cos of phase angle Sun-Moon-Observer
    if (fabs(dist) <= 1.0) dist = acos(dist) / degrad;
    else dist = 180.0;
    if (dist <= 61.0) dist = -3.475256769e-2 + 2.75256769e-2*dist;
    else
     {
      if (dist < 115.0) dist = 0.6962632 * exp(1.48709985e-2*dist);
      else
       {
        if (dist < 155.0) dist = 0.6531068 * exp(1.49213e-2*dist);
        else dist = 1.00779e-9 * pow (dist, 4.486359);
       };
     };
    if (mnmag <= 0) mnmag = 1e-30;  // should never happen.
    mnmag = 0.23 + 5.0*log10(mnmag) + dist;  // moon's mag

    s = carpol(rg);
    dist = s[0];
    hpar = asin (1.0 / dist);  // horizontal parallax
    elev = s[2] / degrad;
    azim = s[1] / degrad;
    azim /= 15.0;
    if (azim < 0) azim += 24.0;
    dms (azim, d, m, hour);
    sprintf
      (otxt, "\n\nMOON:\n\nR.A. %2ih %2im %2.0fs   ", d, m, hour);
    decdot0(otxt);
    dms (elev, d, m, hour);
    sprintf (mtxt, "Dec. %2id %2im %2.0fs   ", d, m, hour);
    decdot0(mtxt);
    strcat(otxt, mtxt);

    AppPos (jd, ep2, lat, lng, indata->geoheight, 1, rm, azim, elev, dist);
    azim /= degrad;
    if (azim < 0) azim += 360.0;
    if (elev > -3.5e-2) elr = (elev + Refract(elev)) / degrad;
    else elr = elev / degrad;
    sprintf (mtxt, "    Azimuth%6.1lf   Elev.%6.1lf \r\n", azim, elr);
    decdot0(mtxt);
    strcat(otxt, mtxt);

    // get rise and set
    hpar = 0.7275 * hpar/degrad - 0.57;
    d = RiseSet (jd, rm, hpar, elev, ep2, lat, indata->geolong, jdr, jds);
    elsv = elev;

    if ((d == 1) && (jdr > jd))
     {
      jd -= 1.0;
      rm = MnPos (indata, jd, ep2, ps);
      d = RiseSet (jd, rm, hpar,
                   elev, ep2, lat, indata->geolong, jdr, jds);
     };
    if (d == 1)   // get dates more accurately
     {
      jd2 = 6.944444444444e-4;  // 1 min time step
      jd1 = jdr - 1.5 / 24.0;
      rfnd = 1;
      for (m = 0; m < 180; m++)
       {
        if (rfnd)
         {
          jd1 += jd2;
          rm = MnPos (indata, jd1, ep2, ps);
          AppPos (jd1, ep2, lat, lng, indata->geoheight, 1, rm,
                            azim, elev, dist);
          if ((elev/degrad) > -0.83)
           {
            if (m == 0) jd1 = jdr - 0.25 / 24.0;  // exotic case
            else
             {
              jdr = jd1;
              rfnd = 0;
             };
           };
         };
       };

      jd1 = jds - 1.5 / 24.0;
      if (jd1 <= jdr) jd1 = jdr + 2*jd2;  // in case we overshot
      rfnd = 1;
      for (m = 0; m < 240; m++)
       {
        if (rfnd)
         {
          jd1 += jd2;
          rm = MnPos (indata, jd1, ep2, ps);
          AppPos (jd1, ep2, lat, lng, indata->geoheight, 1, rm,
                            azim, elev, dist);
          if ((elev/degrad) < -0.83)
           {
            jds = jd1 - jd2;
            rfnd = 0;
           };
         };
       };

      // final check to make sure it's not circumpolar
      d = RiseSet (jdr, rm, hpar,
                   elsv, ep2, lat, indata->geolong, jd1, jd2);
      if (d == 1) d = RiseSet (jd2, rm, hpar,
                               elsv, ep2, lat, indata->geolong, jd1, jd2);
     };

    if (d == 1)
     {
      strcat (otxt, "Rise: ");
      jdr += indata->tzone/24.0;
      dtmstr (jdr, mtxt);
      mtxt[12]= '\0';
      strcat (otxt, mtxt);
      strcat (otxt, "     Set: ");
      jds += indata->tzone/24.0;
      dtmstr (jds, mtxt);
      mtxt[12]= '\0';
      strcat (otxt, mtxt);
     }
    else
     {
      if (d < 0) strcat (otxt, "Moon always below the horizon");
      else strcat (otxt, "Moon always above the horizon");
     };

    // illuminated fraction
    rs = snc - mnc;
    rs = vnorm(rs);
    rm = vnorm(mnc);
    t = (1.0 - dot(rs,rm)) * 0.5;
//    sprintf(mtxt, "\n\nIlluminated Fraction of Moon's disk:%5.2lf",t);
    ps = elm - els;
    if (ps < 0) ps += 2*M_PI;
    ps = ps / (2.0*M_PI);
    sprintf(mtxt, "\n\nIlluminated disk:%5.2lf     Phase:%6.3lf",t, ps);
    decdot0(mtxt);
    strcat(otxt, mtxt);
    sprintf(mtxt, "\nLibration Angles:  L:%7.2lf    B:%7.2lf",lblon, lblat);
    decdot0(mtxt);
    strcat(otxt, mtxt);
    sprintf(mtxt, "\nTerminator:%7.1lf      App.Mag:%7.1lf", termt, mnmag);
    decdot0(mtxt);
    strcat(otxt, mtxt);

    rs = vnorm(sg);
    rg = vnorm(rg);
    t = dot (rs, rg);
    if (fabs(t) <= 1.0) t = acos (t);
    else t = 0;
    t /= degrad;
    sprintf(mtxt, "\nAngular Separation Sun - Moon:%7.2lf deg",t);
    decdot0(mtxt);
    strcat(otxt, mtxt);
  };
 }

/*------------------------- CompYearTot ---------------------------------*/

void CompYearTot (WinpData* indata, double cang, char* otxt)
 {
  // Compute the yearly total of the Sun and the Moon for a specified
  // location.
  // if cang > 1000.0 calculate solar panel statistics instead
	
  double const degrad = M_PI / 180.0;
  double ae = 23454.77992; // 149597870.0/6378.14 =  1AE -> Earth Radii

  char mtxt[50];
  int w1, w2, w3, w4, w5, w6, y, q, m, d;
  double jd0, jd5, lat, lng;
  double azim, elev, elvm, dist, ep2, ps;
  double cossum, azpanel, elpanel, hhx;
  int numsums, spnm1, spnm2, ddx, mmx, yyx;
  bool flytot;
  Vec3 rm, rs;

  if (cang > 1000.0) flytot = false;
  else flytot = true;

  // header output, Julian Date and Sidereal Time
  if (flytot) strcpy(otxt, "Yearly Total for ");
  else strcpy(otxt, "Solar Panel Statistics for ");
  strncat(otxt, indata->geoname, 30);
  if (indata->dmsflg)  // deg.fff -> DD.MMSSS
   {
    lat = DegFDms(indata->geolat);
    lng = DegFDms(indata->geolong);
   }
  else
   {
    lat = indata->geolat;
    lng = indata->geolong;
   };
  sprintf(mtxt,"\nLat: %g   Long: %g   height: %g m\n",
                            lat, lng, indata->geoheight);
  decdot0(mtxt);
  strcat(otxt,mtxt);
  sprintf(mtxt, " Year: %4i \n\n", indata->year);
  strcat(otxt, mtxt);

  spnm1 = 1;
  spnm2 = 12;
  if(!flytot)
   {
    sprintf(mtxt,"\nAzimuth:%6.1lf  Elevation: %5.1lf  Efficiency: %6.2lf % \n",
                         indata->spnaz, indata->spnel, indata->spneff);
    decdot0(mtxt);
    strcat(otxt,mtxt);
    spnm1 = indata->spnm1;
    spnm2 = indata->spnm2;
   };

  lat = indata->geolat * degrad;
  lng = indata->geolong * degrad;

  // now the yearly total

  m = indata->year;
  d = 525600;
  if ((m % 4) == 0)  // leap year
   {
    d = 527040; 
    if (((m%100)==0) && ((m%400)!=0) && ((m>1582))) d = 525600;
   };
  if (m == 1582) d = 511200;  
  jd0 = mjd(1,1,indata->year,0.0);
  w1 = 0;
  w2 = 0;
  w3 = 0;
  w4 = 0;
  w5 = 0;
  w6 = 0;
  m = 0;

  cossum = 0;
  numsums = 0;
  azpanel = indata->spnaz * degrad; 
  elpanel = indata->spnel * degrad;
  elvm = 0;   

  for (q=0; q<d; q++)
   {
    y = q;
    jd5 = jd0 + double(y)/1440.0;

    caldat(jd5, ddx, mmx, yyx, hhx);  // only the month is needed

    if ((mmx >= spnm1) && (mmx <= spnm2))  // months for solar panel check
     {    
      if (m == 0)  // only every 30 minutes
       {
        rs = SnPos (indata, jd5, ep2, ps);
        rs *= ae;
        rm = MnPos (indata, jd5, ep2, ps);
       };
      m++;
      if (m > 29) m = 0;
      AppPos (jd5, ep2, lat, lng, indata->geoheight, 1, rs, azim, elev, dist);
      if (elev > 0)  // only needed for solar panel, but it's quick, so do it always
       {
        numsums++;
        elvm = sin(elev)*sin(elpanel) + cos(elev)*cos(elpanel)*cos(azim-azpanel);
        if (elvm > 0) cossum += elvm;
       };
      elev /= degrad;

      if (flytot)
       {
        AppPos (jd5, ep2, lat, lng, indata->geoheight, 1, rm, azim, elvm, dist);
        elvm /= degrad;
       };

      if (cang <= 90.0) 
       {
        if (elev >= cang) w1++;
        if (elvm >= cang) w5++;
       }
      else
       {
        if (elev >= -0.83) w1++;
        if (elev >= -6.0) w2++;
        if (elev >= -12.0) w3++;
        if (elev >= -18.0) w4++;
        if (elvm >= -0.83) w5++;
        if ((elev <= -18.0) && (elvm < 0.0)) w6++;
       };
     }; // check if month is within interval 
   };  // end for q

  jd5 = double(w1) / 60.0;
  if (cang <= 90.0)
   {
    sprintf (mtxt, "\n  Sun higher than %5.1lf deg : %10.2lf hours \n", cang, jd5);
    decdot0(mtxt);
    strcat(otxt, mtxt);
    jd5 = double(w5) / 60.0;
    sprintf (mtxt, "\n  Moon higher than %5.1lf deg: %10.2lf hours \n", cang, jd5);
    decdot0(mtxt);
    strcat(otxt, mtxt);
   }
  else
   {   
    sprintf (mtxt, "\n  Sunshine Hours:       %10.2lf h \n", jd5);
    decdot0(mtxt);
    strcat(otxt, mtxt);

    if (flytot)  // yearly totals
     {
      jd5 = double(w2) / 60.0;
      sprintf (mtxt, "\n  Civil Twilight:       %10.2lf h ", jd5);
      decdot0(mtxt);
      strcat(otxt, mtxt);
      jd5 = double(w3) / 60.0;
      sprintf (mtxt, "\n  Nautical Twilight:    %10.2lf h ", jd5);
      decdot0(mtxt);
      strcat(otxt, mtxt);

      jd5 = double(w4) / 60.0;
      sprintf (mtxt, "\n  Astronomical Twilight:%10.2lf h \n", jd5);
      decdot0(mtxt);
      strcat(otxt, mtxt);
      jd5 = double(w5) / 60.0;
      sprintf (mtxt, "\n\n  Moonshine Hours:  %10.2lf h \n", jd5);
      decdot0(mtxt);
      strcat(otxt, mtxt);
      jd5 = double(w6) / 60.0;
      sprintf (mtxt, "\n\n  Dark sky without Moon: %10.2lf h \n", jd5);
      decdot0(mtxt);
      strcat(otxt, mtxt);
     }
    else  // solar panel statistics
     {
      if (numsums > 0) cossum = cossum / float(numsums);
      else cossum = 0;
      sprintf (mtxt, 
       "\n Sunshine Percentage received by solar panel: %8.2lf % \n",cossum*100.0);
      decdot0(mtxt);
      strcat(otxt, mtxt);
      cossum = cossum * 1.368 * jd5 * indata->spneff / 100.0;
      sprintf (mtxt, "\n Maximum Electricity produced per m^2: %g kWh \n",cossum);
      decdot0(mtxt);
      strcat(otxt, mtxt);

      if ((spnm1 > 1) || (spnm2 < 12))
       {
        sprintf (mtxt, "\n Months Interval from %2i to %2i \n", spnm1, spnm2);
        strcat(otxt, mtxt);
       };
     };
   };
 
 }  

/*********************************************************************/

/*------------------------- getcfg ---------------------------------*/

int getcfg (WinpData* indata, int& swmax)
 {
  /* get the initial configuration from the file LUNACAL.CFG
     and store the saved values for the timezone and the
     geographic location in indata.
     swmax = 1 if window was maximized, 0 otherwise

     RETURN: 1 if reading the cfg file was successful,
             2 if file does not exist, 0 if error reading file
  */
  int fsc, j;
  double lng, lat, ht, tz;
  double vard[2];
  char gname[31];
  FILE* cfle;

  swmax=0;
  cfle = fopen ("./lunacal.cfg","r");
  if (cfle != NULL)
   {
    fsc = 1;
//    fscanf(cfle,"%lf \n%lf \n%lf \n%lf\n",&tz, &lat, &lng, &ht);
//    if (ferror(cfle)) fsc = 0;
    fsc = Fscanf(cfle, 1, vard);
    if (fsc) 
     {
      tz = vard[0];
      fsc = Fscanf(cfle, 1, vard);
      if (feof(cfle)) fsc = 0;
     };
    if (fsc) 
     {
      lat = vard[0];
      fsc = Fscanf(cfle, 1, vard);
      if (feof(cfle)) fsc = 0;
     };
    if (fsc) 
     {
      lng = vard[0];
      fsc = Fscanf(cfle, 1, vard);
      if (feof(cfle)) fsc = 0;
     };
    if (fsc) 
     {
      ht = vard[0];
      if (feof(cfle)) fsc = 0;
     };

    if (fsc)
     {
      if (fgets(gname, 30, cfle) == NULL) fsc = 0;
      else
       {
        j = strlen(gname);
        if (j>0)
         {
          if(gname[j-1] == '\n') gname[j-1] = '\0';
         };
       };
     };
    if (feof(cfle)) fsc = 0;
    if (fsc) fscanf(cfle,"%d",&swmax);
    if (ferror(cfle))
     {
      fsc = 0;
      clearerr(cfle);
     };
    fclose(cfle);
   }
  else fsc = 2;

  if (fsc == 1)
   {
    indata->tzone = tz;
    indata->geolat = lat;
    indata->geolong = lng;
    indata->geoheight = ht;
    strcpy(indata->geoname,gname);
    if (swmax != 1) swmax = 0;
   };

  return fsc;
 }

/*------------------------- putcfg ---------------------------------*/

void putcfg (WinpData* indata, int swmax)
 {
  // Save configuration data to the file LUNACAL.CFG
  // swmax = 1 if window was maximized, 0 otherwise

  char nbuf[70];	
  FILE* cfle;

  cfle = fopen ("./lunacal.cfg","w");
  if (cfle != NULL)
   {
    sprintf(nbuf,"%lf \n%lf \n%lf \n%lf \n",
       indata->tzone, indata->geolat, indata->geolong, indata->geoheight);
    decdot(nbuf);
    fputs(nbuf, cfle);
    fputs(indata->geoname, cfle);
    fprintf(cfle,"\n%d",swmax);
    fclose(cfle);
   }
 }

 /*------------------------- DefTime ---------------------------------*/

void DefTime (WinpData* indata)  // Get System Time and Date
 {
  time_t tt;
  int hh, mm, ss;
  double jd, hr;

  tt = time(NULL);
  jd = 40587.0 + tt/86400.0 + indata->tzone/24.0; // seconds since 1-JAN-1970

  indata->mjd = jd;
  caldat(jd, hh, mm, ss, hr);
  indata->year = ss;
  indata->month = mm;
  indata->day = hh;

  dms(hr, hh, mm, jd);
  indata->hour = hh;
  indata->minute = mm;
  indata->second = int(jd);
  if (indata->del_auto)	indata->del_tdut = DefTdUt(indata->year);
  };

 /*------------------------- lweinp ---------------------------------*/

void lweinp(WinpData* indata, WoutData* outdata)
 {
  /* display present input parameters */
  char outbuf[81];

  sprintf(outbuf, "The present input values are:");
  PutNxtLn (outdata, outbuf);
  PutSpace(outdata);
  sprintf(outbuf,"Year: \t%4i",indata->year);
  PutNxtLn (outdata, outbuf);
  sprintf(outbuf,"Timezone: %g",indata->tzone);
  decdot(outbuf);
  PutNxtLn (outdata, outbuf);
  if (indata->del_auto)sprintf (outbuf,"Delta-Time set to automatic");
  else sprintf (outbuf,"Delta-Time set to manual");
  PutNxtLn(outdata, outbuf);
  PutSpace(outdata);
  sprintf (outbuf,"Geographic Location:");
  PutNxtLn(outdata, outbuf);
  PutNxtLn (outdata, indata->geoname);
  sprintf(outbuf,"Latitude:  \t%g", indata->geolat);
  decdot(outbuf);
  PutNxtLn (outdata, outbuf);
  sprintf(outbuf,"Longitude: \t%g", indata->geolong);
  decdot(outbuf);
  PutNxtLn (outdata, outbuf);
  sprintf(outbuf,"Height:    \t%g", indata->geoheight);
  decdot(outbuf);
  PutNxtLn (outdata, outbuf);
 }

 /*------------------------- lwelcm ---------------------------------*/

void lwelcm(WinpData* indata, WoutData* outdata)
 {
  // prepare welcoming message and display initial input values
  
  int swmax;
  char outbuf[81];

  // default data for indata
  indata->tzone = 0;
  indata->del_auto = 1;
  DefTime(indata);
  indata->time = 0;
  indata->tstep = 300.0;
  indata->del_tdut = DefTdUt(indata->year);
  indata->tmfirst = 1;
  indata->svfapnd = 1;
  indata->svfirst = 1;
  indata->geolat = 0;
  indata->geolong = 0;
  indata->geoheight = 0;
  indata->eclheight = 35788.1189;
  indata->geoid = false;
  strcpy(indata->geoname,"Default");
  strcpy(indata->svfname,"./lunasave.txt");
  indata->dmsflg = 0;
  indata->filsflg = 0;
  indata->lstcall = 0;
  indata->invalflg = 0;
  indata->locecl = 0;
  indata->eclbuf.lastyear = -9999;
  indata->eclbuf.numecl = 0;
  indata->eclbuf.lasttz = indata->tzone;
  indata->eclbuf.lastdlt = indata->del_tdut;
  indata->eclbuf.cstep = 30;
  indata->eclbuf.emagrid = 0;
  indata->eclbuf.emagrey = 1;
  indata->eclbuf.cmxlat = 0;
  indata->eclbuf.cmxlng = 0;
  indata->mapprj = 0;
  indata->centmer = 0.0;
  indata->spnaz = 180.0;
  indata->spnel = 45.0;
  indata->spneff = 15.0;
  indata->spnm1 = 1;
  indata->spnm2 = 12;

  sprintf (outbuf, "     *** Welcome to LunaCal Version 5.6 ***");
  PutNxtLn (outdata, outbuf);
  PutSpace(outdata);

  // read the configuration data
  if (getcfg (indata, swmax) < 1)
   {
    sprintf (outbuf,
       "Problems reading Configuration File. Using defaults instead.");
    PutNxtLn (outdata, outbuf);
    PutSpace(outdata);
   }
  else DefTime(indata);

  lweinp(indata, outdata);
 }


